import asyncio
from asyncio import Queue, create_task

from melobot.bot import Bot
from melobot.handle import GetParseArgs, on_start_match
from melobot.log import GenericLogger
from melobot.plugin import PluginPlanner
from melobot.protocols.onebot.v11.adapter.base import Adapter
from melobot.protocols.onebot.v11.adapter.event import MessageEvent
from melobot.protocols.onebot.v11.io.base import BaseIOSource
from melobot.protocols.onebot.v11.io.packet import EchoPacket, InPacket, OutPacket
from melobot.protocols.onebot.v11.utils import GroupMsgChecker, LevelRole
from melobot.utils.parse import CmdArgs, CmdParser
from tests.base import *

_GRUOP_EVENT_DICT = {
    "time": 1725292489,
    "self_id": 123456,
    "post_type": "message",
    "message_type": "group",
    "sub_type": "normal",
    "sender": {
        "age": 0,
        "nickname": "这是一个群昵称",
        "sex": "unknown",
        "user_id": 3,
        "area": "",
        "card": "",
        "level": "",
        "role": "member",
        "title": "",
    },
    "message_id": -1234567890,
    "font": 0,
    "message": "\n\r\t .echo/123/456    \r\n",
    "user_id": 3,
    "anonymous": None,
    "group_id": 6,
    "raw_message": "45691237\n\r\t .echo/123/456    \r\n",
}


h = on_start_match(
    ["123", "456"],
    checker=GroupMsgChecker(
        role=LevelRole.WHITE,
        owner=1,
        super_users=[2, 3],
        white_users=[4],
        black_users=[5],
        white_groups=[6, 7],
    ),
    parser=CmdParser(".", "/", "echo"),
)


@h
async def _flow(
    bot: Bot,
    event: MessageEvent,
    logger: GenericLogger,
    args: CmdArgs = GetParseArgs(),
) -> None:
    logger.info(args)
    await bot.close()
    _SUCCESS_SIGNAL.set()


class TempIO(BaseIOSource):
    def __init__(self) -> None:
        super().__init__(1)
        self.queue = Queue()
        self.queue.put_nowait(InPacket(data=_GRUOP_EVENT_DICT))

    async def open(self) -> None:
        pass

    def opened(self) -> bool:
        return True

    async def close(self) -> None:
        pass

    async def input(self) -> InPacket:
        return await self.queue.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        if packet.echo_id is None:
            return EchoPacket(noecho=True)


_SUCCESS_SIGNAL = asyncio.Event()


async def test_adapter_base():
    mbot = Bot("test_handle")
    mbot.add_io(TempIO())
    mbot.add_adapter(Adapter())
    mbot.load_plugin(PluginPlanner("1.0.0", flows=[_flow]))
    create_task(mbot.core_run())
    await mbot._rip_signal.wait()
    await _SUCCESS_SIGNAL.wait()



# -*- encoding: utf-8 -*-
# @File   : test_utils.py
# @Time   : 2024/08/26 20:53:04
# @Author : Kariko Lin

import asyncio
from enum import Enum
from random import choice, randint

from typing_extensions import Any, Coroutine

from melobot.utils import *
from tests.base import *


async def test_singleton() -> None:
    @singleton
    class ChlorideP:
        def __init__(self, sign: Any) -> None:
            self.sign = sign

        def __str__(self) -> str:
            return f"Chloride with sign: {self.sign}"

    a, b = ChlorideP("nya"), ChlorideP("rua")
    assert a is b and b.sign == a.sign


# Interesting names :)
# Try to search them in some places
ESTHER_EGGS = [
    "PedroDelMar",
    "ElapsingDreams",
    "ShimakazeProject",
    "MelorenAe",
    "aiCoreIn",
    "MelodyEcho",
    "SnowyKami",
    "LiteeCiallo",
    "DislinkSforza",
]


class TestRwc:
    RWC = RWContext(10)
    EGGS = ESTHER_EGGS.copy()
    BUF = EGGS.copy()
    READING_FLAGS = [False for _ in range(len(EGGS))]
    ASYNC_READED = False

    @classmethod
    async def write(cls) -> None:
        async with cls.RWC.write():
            i = randint(0, len(TestRwc.EGGS) - 1)
            val = cls.BUF[i]
            if val == "":
                raise ValueError("写写冲突")

            cls.BUF[i] = ""
            await aio.sleep(0.01)
            cls.BUF[i] = val

    @classmethod
    async def read(cls) -> None:
        async with cls.RWC.read():
            i = randint(0, len(TestRwc.EGGS))
            val = cls.BUF[i]
            if val == "":
                raise ValueError("读写冲突")
            if cls.READING_FLAGS[i]:
                cls.ASYNC_READED = True

            cls.READING_FLAGS[i] = True
            await aio.sleep(0.01)
            cls.READING_FLAGS[i] = False

    @classmethod
    async def test_rwcontrol(cls) -> None:
        # Ensure concurrent reading be encountered in task seqs
        r_seq = [cls.read() for _ in range(5)]

        # Avoid GC complaint that caused by unclosed coroutine
        getters = [cls.read, cls.write]
        rw_seq = [getters[choice([0, 1])]() for _ in range(50)]

        idx = choice(range(0, len(r_seq) + len(rw_seq) - 1))
        seq = rw_seq[:idx] + r_seq + rw_seq[idx:]
        await aio.wait(map(aio.create_task, seq))
        assert cls.ASYNC_READED


async def test_get_id() -> None:
    n = 100000
    ids = [get_id() for _ in range(n)]
    assert n - len(set(ids)) <= 1


class TestAsyncInterfaceAdapter:
    async def f1() -> None: ...

    def f2() -> int:
        return 1

    async def f3() -> int:
        return 1

    @classmethod
    def f4(cls) -> Coroutine[None, None, int]:
        return cls.f3()

    f5 = lambda: TestAsyncInterfaceAdapter.f3()

    async def f6(x: int, y: int) -> int:
        return x + y

    @classmethod
    async def test_to_async(cls) -> None:
        assert to_async(cls.f1) is cls.f1

        ret = await to_async(cls.f2)()
        assert ret == 1

        ret = await to_async(cls.f3)()
        assert ret == 1

        ret = await to_async(cls.f4)()
        assert ret == 1

        ret = await to_async(cls.f5)()
        assert ret == 1

    @classmethod
    async def test_to_coro(cls) -> None:
        coro = cls.f1()
        assert to_coro(coro) is coro
        coro.close()

        ret = await to_coro(cls.f2)
        assert ret == 1

        ret = await to_coro(cls.f3)
        assert ret == 1

        ret = await to_coro(cls.f4)
        assert ret == 1

        ret = await to_coro(cls.f5)
        assert ret == 1

        ret = await to_coro(cls.f6, 1, y=2)
        assert ret == 3


class TestIfNot:
    REJECTED = False
    RET = 0

    @classmethod
    def restore(cls) -> None:
        cls.REJECTED = False
        cls.RET = 0

    @classmethod
    async def reject(cls) -> None:
        cls.REJECTED = True

    @staticmethod
    async def get_cond() -> bool:
        return False

    @classmethod
    async def test_if_not(cls) -> None:
        async def func() -> None:
            cls.RET = 1

        f1 = if_not(condition=cls.get_cond, reject=cls.reject, give_up=True)(func)
        await f1()
        assert cls.REJECTED
        assert cls.RET == 0

        cls.restore()

        f2 = if_not(condition=cls.get_cond, reject=cls.reject, give_up=False)(func)
        await f2()
        assert cls.REJECTED
        assert cls.RET == 1


class TestUnfoldCtx:
    VAL = 0

    class SyncCtx:
        def __enter__(self) -> None:
            TestUnfoldCtx.VAL = 1

        def __exit__(self, *_, **__) -> None:
            TestUnfoldCtx.VAL = 0

    class AsyncCtx:
        async def __aenter__(self) -> None:
            TestUnfoldCtx.VAL = 1

        async def __aexit__(self, *_, **__) -> None:
            TestUnfoldCtx.VAL = 0

    @classmethod
    async def test_unfold_ctx(cls) -> None:
        async def func() -> None:
            assert cls.VAL == 1

        f = unfold_ctx(cls.SyncCtx)(func)
        await f()
        assert cls.VAL == 0
        f = unfold_ctx(cls.AsyncCtx)(func)
        await f()
        assert cls.VAL == 0


class Status(Enum):
    TOO_MANY_REQUESTS = 1
    REQUEST_TIMEOUT = 2
    EXPECTATION_FAILED = 3
    OK = 4


class TestAsyncUtils:
    NYA = ""

    async def lock_callback(self) -> int:
        return Status.TOO_MANY_REQUESTS

    async def time_callback(self) -> int:
        return Status.REQUEST_TIMEOUT

    async def cd_callback(self, timeout: float) -> int:
        return Status.TOO_MANY_REQUESTS

    async def test_lock(self) -> None:
        @lock(self.lock_callback)
        async def gen_ini_string() -> int:  # using NYA
            if not TestAsyncUtils.NYA:
                TestAsyncUtils.NYA += f"[{choice(ESTHER_EGGS)}]\n"
            await aio.sleep(0)
            TestAsyncUtils.NYA += f"clsid = {get_id()}"
            return Status.OK

        coropool = [gen_ini_string() for _ in ESTHER_EGGS]
        rets = await aio.gather(*coropool)
        assert rets.count(Status.OK) == 1

    async def test_cooldown(self) -> None:
        @cooldown(cd_callback=self.cd_callback, interval=3)
        async def sendmsg() -> int:
            # print(f"I'm searching {choice(ESTHER_EGGS)} in melobot group.")
            return Status.OK

        pool = [sendmsg() for _ in range(5)]
        rets = await aio.gather(*pool)
        assert rets.count(Status.TOO_MANY_REQUESTS) > 0

    async def test_semaphore(self) -> None:
        tickets = 10

        @semaphore(value=1)
        async def buyticket() -> int:
            nonlocal tickets
            if tickets <= 0:
                return Status.EXPECTATION_FAILED
            tickets -= 1
            return Status.OK

        pool = [buyticket() for _ in range(11)]
        rets = await aio.gather(*pool)
        assert not tickets and Status.EXPECTATION_FAILED in rets

    async def test_timelimit(self) -> None:
        @timelimit(self.time_callback, timeout=0.25)
        async def foo() -> int:
            await aio.sleep(0.5)
            return Status.OK

        ret = await foo()
        assert (await foo()) == Status.REQUEST_TIMEOUT

    async def test_speedlimit(self) -> None:
        """In my opinion, just name it 'chancelimit'."""

        @speedlimit(self.lock_callback, limit=1, duration=5)
        async def foo() -> int:
            await aio.sleep(0)
            return Status.OK

        pool = [foo() for _ in range(10)]
        rets = await aio.gather(*pool)
        assert rets.count(Status.OK) <= 1


class TimeGetter:
    def __get__(self, *_, **__) -> float:
        return aio.get_event_loop().time()


class TestCallableDispatch:
    # Use a descriptor as dynamic class var
    time = TimeGetter()
    TEST_ATTR = "__only_for_test__"

    @staticmethod
    def foo(e: aio.Event) -> None:
        e.set()

    @staticmethod
    async def afoo(e: aio.Event) -> None:
        await aio.sleep(0)
        e.set()

    @classmethod
    async def abar(cls, obj: Any) -> None:
        if not hasattr(obj, cls.TEST_ATTR):
            setattr(obj, cls.TEST_ATTR, 0)
        setattr(obj, cls.TEST_ATTR, getattr(obj, cls.TEST_ATTR) + 1)

    @classmethod
    async def test_call_later(cls) -> None:
        e = aio.Event()
        call_later(lambda: cls.foo(e), 0.1)
        begin = cls.time
        await e.wait()
        assert cls.time - begin <= 0.2

    @classmethod
    async def test_call_at(cls) -> None:
        e = aio.Event()
        call_at(lambda: cls.foo(e), cls.time + 0.1)
        begin = cls.time
        await e.wait()
        assert cls.time - begin <= 0.2

    @classmethod
    async def test_async_later(cls) -> None:
        e = aio.Event()
        await async_later(cls.afoo(e), 0.1)
        begin = cls.time
        await e.wait()
        assert cls.time - begin <= 0.1

    @classmethod
    async def test_async_at(cls) -> None:
        e = aio.Event()
        await async_at(cls.afoo(e), cls.time + 0.1)
        begin = cls.time
        await e.wait()
        assert cls.time - begin <= 0.1

    @classmethod
    async def test_async_interval(cls) -> None:
        obj = type("__XXX", (object,), {})()
        t = async_interval(lambda: cls.abar(obj), 0.1)
        await aio.sleep(0.5)
        t.cancel()
        try:
            await t
        except asyncio.CancelledError:
            pass
        assert getattr(obj, cls.TEST_ATTR) >= 3



from abc import abstractmethod

from melobot.io import (
    AbstractInSource,
    AbstractIOSource,
    AbstractOutSource,
    AbstractSource,
)

from ..const import PROTOCOL_IDENTIFIER
from .packet import EchoPacket, InPacket, OutPacket


class BaseSource(AbstractSource):
    def __init__(self) -> None:
        super().__init__()
        self.protocol = PROTOCOL_IDENTIFIER
        self._hook_bus.set_tag(f"{self.protocol}/{self.__class__.__name__}")

    @abstractmethod
    async def open(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def opened(self) -> bool:
        raise NotImplementedError


class BaseInSource(AbstractInSource[InPacket], BaseSource):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def open(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def opened(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def input(self) -> InPacket:
        raise NotImplementedError


class BaseOutSource(AbstractOutSource[OutPacket, EchoPacket], BaseSource):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    async def open(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def opened(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def output(self, packet: OutPacket) -> EchoPacket:
        raise NotImplementedError


class BaseIOSource(
    AbstractIOSource[InPacket, OutPacket, EchoPacket], BaseInSource, BaseOutSource
):
    """

    :ivar float cd_time: 发送行为操作的冷却时间（防风控）
    """

    # pylint: disable=duplicate-code
    def __init__(self, cd_time: float) -> None:
        super().__init__()
        self.cd_time = cd_time if cd_time >= 0 else 0

    @abstractmethod
    async def open(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def opened(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    async def input(self) -> InPacket:
        raise NotImplementedError

    @abstractmethod
    async def output(self, packet: OutPacket) -> EchoPacket:
        raise NotImplementedError



from .action import (
    Action,
    CanSendImageAction,
    CanSendRecordAction,
    CleanCacheAction,
    DeleteMsgAction,
    GetCookiesAction,
    GetCredentialsAction,
    GetCsrfTokenAction,
    GetForwardMsgAction,
    GetFriendlistAction,
    GetGroupHonorInfoAction,
    GetGroupInfoAction,
    GetGrouplistAction,
    GetGroupMemberInfoAction,
    GetGroupMemberlistAction,
    GetImageAction,
    GetLoginInfoAction,
    GetMsgAction,
    GetRecordAction,
    GetStatusAction,
    GetStrangerInfoAction,
    GetVersionInfoAction,
    SendForwardMsgAction,
    SendLikeAction,
    SendMsgAction,
    SetFriendAddRequestAction,
    SetGroupAddRequestAction,
    SetGroupAdminAction,
    SetGroupAnonymousAction,
    SetGroupAnonymousBanAction,
    SetGroupBanAction,
    SetGroupCardAction,
    SetGroupKickAction,
    SetGroupLeaveAction,
    SetGroupNameAction,
    SetGroupSpecialTitleAction,
    SetGroupWholeBanAction,
    SetRestartAction,
    msgs_to_dicts,
)
from .base import Adapter, EchoFactory, EchoRequireCtx, EventFactory, OutputFactory
from .echo import (
    CanSendImageEcho,
    CanSendRecordEcho,
    Echo,
    EmptyEcho,
    GetCookiesEcho,
    GetCredentialsEcho,
    GetCsrfTokenEcho,
    GetForwardMsgEcho,
    GetFriendListEcho,
    GetGroupHonorInfoEcho,
    GetGroupInfoEcho,
    GetGroupListEcho,
    GetGroupMemberInfoEcho,
    GetGroupMemberListEcho,
    GetImageEcho,
    GetLoginInfoEcho,
    GetMsgEcho,
    GetRecordEcho,
    GetStatusEcho,
    GetStrangerInfoEcho,
    GetVersionInfoEcho,
    SendForwardMsgEcho,
    SendMsgEcho,
)
from .event import (
    Event,
    FriendAddNoticeEvent,
    FriendRecallNoticeEvent,
    FriendRequestEvent,
    GroupAdminNoticeEvent,
    GroupBanNoticeEvent,
    GroupDecreaseNoticeEvent,
    GroupIncreaseNoticeEvent,
    GroupMessageEvent,
    GroupRecallNoticeEvent,
    GroupRequestEvent,
    GroupUploadNoticeEvent,
    HeartBeatMetaEvent,
    HonorNotifyEvent,
    LifeCycleMetaEvent,
    LuckyKingNotifyEvent,
    MessageEvent,
    MetaEvent,
    NoticeEvent,
    NotifyNoticeEvent,
    PokeNotifyEvent,
    PrivateMessageEvent,
    RequestEvent,
)
from .segment import (
    AnonymousSegment,
    AtSegment,
    ContactFriendSegment,
    ContactGroupSegment,
    ContactSegment,
    DiceSegment,
    FaceSegment,
    ForwardSegment,
    ImageRecvSegment,
    ImageSegment,
    ImageSendSegment,
    JsonSegment,
    LocationSegment,
    MusicCustomSegment,
    MusicPlatformSegment,
    MusicSegment,
    NodeGocqCustomSegment,
    NodeReferSegment,
    NodeSegment,
    NodeStdCustomSegment,
    PokeRecvSegment,
    PokeSegment,
    PokeSendSegment,
    RecordRecvSegment,
    RecordSegment,
    RecordSendSegment,
    ReplySegment,
    RpsSegment,
    Segment,
    ShakeSegment,
    ShareSegment,
    TextSegment,
    VideoRecvSegment,
    VideoSegment,
    VideoSendSegment,
    XmlSegment,
    base64_encode,
    contents_to_segs,
    cq_anti_escape,
    cq_escape,
    cq_filter_text,
    segs_to_contents,
)



from __future__ import annotations

from typing_extensions import Any, ClassVar, Generic, Literal, NamedTuple, NoReturn

from .typ.base import T

__version__ = "3.1.1"


def _version_str_to_info(s: str) -> VersionInfo:
    parts = s.split(".")
    major = int(parts[0])
    minor = int(parts[1])

    remains = parts[2]
    flag: Literal["alpha", "beta", "pre-release", "final"]

    if "a" in remains:
        _parts = remains.split("a")
        flag = "alpha"
    elif "b" in remains:
        _parts = remains.split("b")
        flag = "beta"
    elif "rc" in remains:
        _parts = remains.split("rc")
        flag = "pre-release"
    else:
        _parts = [remains, remains]
        flag = "final"

    micro = int(_parts[0])
    serial = int(_parts[1])
    return VersionInfo(major, minor, micro, flag, serial)


class ReadOnlyAttr(Generic[T]):
    def __init__(self, val: T):
        self.val = val

    def __get__(self, obj: Any, klass: Any = None) -> T:
        return self.val

    def __set__(self, obj: Any, value: T) -> NoReturn:
        raise AttributeError("只读属性无法重新设定值")


class VersionInfo(NamedTuple):
    """版本信息元组"""

    major: int
    """主要版本号

    变更时伴随重大架构更新（下一代 melobot 才会更改）
    """

    minor: int
    """次要版本号

    变更时伴随重要的功能更新
    """

    micro: int
    """微版本号

    变更时伴随不重要的功能更新，或是漏洞/错误修复
    """

    releaselevel: Literal["alpha", "beta", "pre-release", "final"]
    """发行级别

    `micro` 更新，一般直接使用 `final` 发行级别

    `minor` 更新，可能历经：
        - `alpha` -> `beta` -> `pre-release` -> `final` （极少）
        - `pre-release` -> `final` （较多）
        - `final` （少）

    `major` 更新，一定历经：
        - `alpha` -> `beta` -> `pre-release` -> `final`
    """

    serial: int
    """发行序列号，若 `releaselevel` 为 "final"，则该值与 `micro` 相同

    否则可以视为具体 `releaselevel` 的子版本号，相当于 `nano` 级版本号
    """


class MetaInfoMeta(type):
    ver = ReadOnlyAttr[str](__version__)
    ver_info = ReadOnlyAttr[VersionInfo](_version_str_to_info(__version__))
    name = ReadOnlyAttr[str]("melobot")
    desc = ReadOnlyAttr[str]("A bot framework with much high level features.")
    src = ReadOnlyAttr[str]("https://github.com/Meloland/melobot")
    logo = ReadOnlyAttr[str](
        "\n".join(
            (
                r"                _       _           _   ",
                r" _ __ ___   ___| | ___ | |__   ___ | |_ ",
                r"| '_ ` _ \ / _ \ |/ _ \| '_ \ / _ \| __|",
                r"| | | | | |  __/ | (_) | |_) | (_) | |_ ",
                r"|_| |_| |_|\___|_|\___/|_.__/ \___/ \__|",
            )
        )
    )


class MetaInfo(metaclass=MetaInfoMeta):
    """melobot 项目只读元信息"""

    ver: ClassVar[str]
    """melobot 版本

       :meta hide-value:
    """

    ver_info: ClassVar[VersionInfo]
    """melobot 版本信息

       :meta hide-value:
    """

    name: ClassVar[str]
    """melobot 项目名称

       :meta hide-value:
    """

    desc: ClassVar[str]
    """melobot 项目描述

       :meta hide-value:
    """

    src: ClassVar[str]
    """melobot 项目地址

       :meta hide-value:
    """

    logo: ClassVar[str]
    """melobot ascii art 图标

       :meta hide-value:
    """



from asyncio import Lock
from functools import partial, wraps

from typing_extensions import Callable, Sequence, cast

from ..adapter.model import Event, TextEvent
from ..ctx import FlowCtx, ParseArgsCtx
from ..di import Depends, inject_deps
from ..session.base import enter_session
from ..session.option import DefaultRule, Rule
from ..typ._enum import LogicMode
from ..typ.base import AsyncCallable, SyncOrAsyncCallable
from ..utils.check import Checker, checker_join
from ..utils.common import get_obj_name
from ..utils.match import (
    ContainMatcher,
    EndMatcher,
    FullMatcher,
    Matcher,
    RegexMatcher,
    StartMatcher,
)
from ..utils.parse import AbstractParseArgs, CmdArgFormatter, CmdParser, Parser
from .base import Flow, no_deps_node


def GetParseArgs() -> AbstractParseArgs:  # pylint: disable=invalid-name
    """获取解析参数

    :return: 解析参数
    """
    return cast(AbstractParseArgs, Depends(ParseArgsCtx().get, recursive=False))


class FlowDecorator:
    def __init__(
        self,
        checker: Checker | None | Callable[[Event], bool] = None,
        matcher: Matcher | None = None,
        parser: Parser | None = None,
        priority: int = 0,
        block: bool = False,
        temp: bool = False,
        decos: Sequence[Callable[[Callable], Callable]] | None = None,
        rule: Rule[Event] | None = None,
    ) -> None:
        self.checker: Checker | None
        if callable(checker):
            self.checker = Checker.new(checker)
        else:
            self.checker = checker

        self.matcher = matcher
        self.parser = parser
        self.priority = priority
        self.block = block
        self.decos = decos
        self.rule = rule

        self._temp = temp
        self._invalid = False
        self._lock = Lock()
        self._flow: Flow

    async def _pre_process(self, event: Event) -> tuple[bool, AbstractParseArgs | None]:
        if self.checker:
            status = await self.checker.check(event)
            if not status:
                return (False, None)

        args: AbstractParseArgs | None = None
        if isinstance(event, TextEvent):
            if self.matcher:
                status = await self.matcher.match(event.text)
                if not status:
                    return (False, None)

            if self.parser:
                args = await self.parser.parse(event.text)
                if args:
                    return (True, args)
                return (False, None)

        return (True, None)

    async def _process(
        self,
        func: AsyncCallable[..., bool | None],
        event: Event,
        args: AbstractParseArgs | None,
    ) -> bool | None:
        event.spread = not self.block
        parse_args_ctx = ParseArgsCtx()
        if args is not None:
            args_token = parse_args_ctx.add(args)
        else:
            args_token = None

        try:
            if self.rule is not None:
                async with enter_session(self.rule):
                    return await func()
            return await func()
        finally:
            if args_token:
                parse_args_ctx.remove(args_token)

    async def auto_flow_wrapped(
        self, func: AsyncCallable[..., bool | None]
    ) -> bool | None:
        if self._invalid:
            self._flow.dismiss()
            return None

        event = FlowCtx().get().completion.event
        if not self._temp:
            passed, args = await self._pre_process(event)
            if not passed:
                return None
            return await self._process(func, event, args)

        async with self._lock:
            if self._invalid:
                self._flow.dismiss()
                return None

            passed, args = await self._pre_process(event)
            if not passed:
                return None
            self._invalid = True

        return await self._process(func, event, args)

    def __call__(self, func: SyncOrAsyncCallable[..., bool | None]) -> Flow:
        func = inject_deps(func)
        if self.decos is not None:
            for deco in reversed(self.decos):
                func = deco(func)

        n = no_deps_node(wraps(func)(partial(self.auto_flow_wrapped, func)))
        n.name = get_obj_name(func, otype="callable")
        self._flow = Flow(f"OneBotV11Flow[{n.name}]", (n,), priority=self.priority)
        return self._flow


def on_event(
    checker: Checker | None | Callable[[Event], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    rule: Rule[Event] | None = None,
) -> FlowDecorator:
    return FlowDecorator(checker, matcher, parser, priority, block, temp, decos, rule)


def on_text(
    checker: Checker | None | Callable[[TextEvent], bool] = None,
    matcher: Matcher | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    if legacy_session:
        rule = DefaultRule()
    else:
        rule = None

    return on_event(
        checker_join(lambda e: isinstance(e, TextEvent), checker),  # type: ignore[arg-type]
        matcher,
        parser,
        priority,
        block,
        temp,
        decos,
        cast(Rule[Event] | None, rule),
    )


def on_command(
    cmd_start: str | list[str],
    cmd_sep: str | list[str],
    targets: str | list[str],
    fmtters: list[CmdArgFormatter | None] | None = None,
    checker: Checker | None | Callable[[TextEvent], bool] = None,
    matcher: Matcher | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        matcher,
        CmdParser(cmd_start=cmd_start, cmd_sep=cmd_sep, targets=targets, fmtters=fmtters),
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )


def on_start_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        StartMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )


def on_contain_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        ContainMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )


def on_full_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        FullMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )


def on_end_match(
    target: str | list[str],
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        EndMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )


def on_regex_match(
    target: str,
    logic_mode: LogicMode = LogicMode.OR,
    checker: Checker | Callable[[TextEvent], bool] | None = None,
    parser: Parser | None = None,
    priority: int = 0,
    block: bool = False,
    temp: bool = False,
    decos: Sequence[Callable[[Callable], Callable]] | None = None,
    legacy_session: bool = False,
) -> FlowDecorator:
    return on_text(
        checker,
        RegexMatcher(target, logic_mode),
        parser,
        priority,
        block,
        temp,
        decos,
        legacy_session,
    )



from typing import Literal

from pydantic import ValidationError
from typing_extensions import TypedDict

from melobot.protocols.onebot.v11.adapter import segment as seg
from tests.base import *


async def test_smart_union():
    s = seg.ImageSegment(file="12345678")
    s = seg.ImageSegment(file="12345678.jpg", url="https://example.com/12345678.jpg")
    s = seg.ContactSegment(type="qq", id=12345678)
    s = seg.ContactSegment(type="group", id=12345678)
    s = seg.MusicSegment(
        type="custom",
        url="https://example.com/music.mp3",
        audio="https://example.com/audio.mp3",
        title="music",
    )


async def test_type_and_data():
    s = seg.Segment("my", key="123")
    assert s.type == "my"
    assert s.data == {"key": "123"}


async def test_add_type():
    SType = seg.Segment.add_type(Literal["MyS"], TypedDict("MyData", {"key": str}))
    s = SType(key="123")
    assert s.type == "MyS"
    assert s.data == {"key": "123"}
    assert s.raw == {"type": "MyS", "data": {"key": "123"}}

    s2 = seg.Segment.resolve("MyS", {"key": "123"})
    assert s.type == "MyS"
    assert s.data == {"key": "123"}
    assert s.raw == {"type": "MyS", "data": {"key": "123"}}
    assert isinstance(s2, SType)

    with pt.raises(ValidationError):
        SType(key=123)


async def test_serialize_and_deserialize():
    sl1 = seg.Segment.__resolve_cq__("12345[CQ:image,file=12345678.jpg]123456")
    sl2 = seg.Segment.__resolve_cq__(
        "123&#91;45[CQ:node,user_id=10001000,nickname=某人,content=&#91;CQ:face&#44;id=123&#93;哈喽～]12345"
    )

    s3 = seg.ImageSegment(file="12345678")
    s4 = seg.RpsSegment()
    s5 = seg.Segment.__resolve_cq__("12345")[0]
    s6 = seg.Segment.__resolve_cq__("[CQ:image,file=12345678.jpg]")[0]
    s7 = seg.AtSegment(15742)
    s8 = seg.Segment.__resolve_cq__("")

    assert s8 == []

    assert sl1[0].to_dict() == {"type": "text", "data": {"text": "12345"}}
    assert sl1[1].to_dict() == {"type": "image", "data": {"file": "12345678.jpg"}}
    assert sl1[2].to_dict() == {"type": "text", "data": {"text": "123456"}}
    assert sl2[0].to_dict() == {"type": "text", "data": {"text": "123[45"}}
    assert sl2[1].to_dict() == {
        "type": "node",
        "data": {
            "user_id": 10001000,
            "nickname": "某人",
            "content": [
                {"type": "face", "data": {"id": 123}},
                {"type": "text", "data": {"text": "哈喽～"}},
            ],
        },
    }
    assert sl2[1].to_dict(force_str=True) == {
        "type": "node",
        "data": {
            "user_id": "10001000",
            "nickname": "某人",
            "content": [
                {"type": "face", "data": {"id": "123"}},
                {"type": "text", "data": {"text": "哈喽～"}},
            ],
        },
    }
    assert s7.to_dict(force_str=True) == {"type": "at", "data": {"qq": "15742"}}
    assert sl2[2].to_dict() == {"type": "text", "data": {"text": "12345"}}
    assert s5.to_dict() == {"type": "text", "data": {"text": "12345"}}
    assert s6.to_dict() == {"type": "image", "data": {"file": "12345678.jpg"}}

    assert s3.to_cq() == "[CQ:image,file=12345678]"
    assert s4.to_cq() == "[CQ:rps]"
    assert "".join(s.to_cq() for s in sl1) == "12345[CQ:image,file=12345678.jpg]123456"
    assert (
        sl2[0].to_cq(escape=True) + sl2[1].to_cq() + sl2[2].to_cq()
        == "123&#91;45[CQ:node,user_id=10001000,nickname=某人,content=&#91;CQ:face&#44;id=123&#93;哈喽～]12345"
    )
    assert s5.to_cq() == "12345"
    assert s6.to_cq() == "[CQ:image,file=12345678.jpg]"

    assert s6.to_json() == '{"type": "image", "data": {"file": "12345678.jpg"}}'



from __future__ import annotations

import asyncio
import base64
import importlib
import time
import warnings
from contextlib import asynccontextmanager
from datetime import datetime
from functools import wraps

from typing_extensions import Any, AsyncGenerator, Callable, Literal, cast

from ..typ.base import P, T


def get_obj_name(
    obj: Any,
    otype: Literal["callable", "class", "object"] | str = "object",
    default: str = "<anonymous %s>",
) -> str:
    """获取一个对象的限定名称或名称，这适用于一些类型较宽的参数。

    无法获取有效名称时，产生一个 `default % otype` 字符串

    例如某处接受一个 `Callable` 类型的参数，对于一般函数来说，使用
    `__qualname__` 或 `__name__` 可获得名称，但某些可调用对象这些值可能为 `None`
    或不存在。使用此方法可保证一定返回字符串

    .. code:: python

        def _(a: Callable) -> None:
            valid_str: str = get_obj_name(a, otype="callable")

        def _(a: type) -> None:
            valid_str: str = get_obj_name(a, otype="class")

        def _(a: Any) -> None:
            valid_str: str = get_obj_name(a, otype="type of a, only for str concat")


    :param obj: 对象
    :param otype: 预期的对象类型
    :param default: 无法获取任何有效名称时的默认字符串
    :return: 对象名称或默认字符串
    """
    if hasattr(obj, "__qualname__"):
        return cast(str, obj.__qualname__)

    if hasattr(obj, "__name__"):
        return cast(str, obj.__name__)

    return default % otype


def singleton(cls: Callable[P, T]) -> Callable[P, T]:
    """单例装饰器

    :param cls: 需要被单例化的可调用对象
    :return: 需要被单例化的可调用对象
    """
    obj_map = {}

    @wraps(cls)
    def singleton_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
        if cls not in obj_map:
            obj_map[cls] = cls(*args, **kwargs)
        return obj_map[cls]

    return singleton_wrapped


def deprecate_warn(msg: str, stacklevel: int = 2) -> None:
    # pylint: disable=cyclic-import
    from ..ctx import LoggerCtx

    if logger := LoggerCtx().try_get():
        logger.warning(msg)
    warnings.simplefilter("always", DeprecationWarning)
    warnings.warn(msg, category=DeprecationWarning, stacklevel=stacklevel)
    warnings.simplefilter("default", DeprecationWarning)


def deprecated(msg: str) -> Callable[[Callable[P, T]], Callable[P, T]]:

    def deprecated_wrapper(func: Callable[P, T]) -> Callable[P, T]:

        @wraps(func)
        def deprecated_wrapped(*args: P.args, **kwargs: P.kwargs) -> T:
            deprecate_warn(
                f"使用了弃用函数/方法 {func.__module__}.{func.__qualname__}: {msg}",
                stacklevel=3,
            )
            return func(*args, **kwargs)

        return deprecated_wrapped

    return deprecated_wrapper


class DeprecatedLoader:
    def __init__(self, mod_name: str, obj_pairs: dict[str, tuple[str, str, str]]) -> None:
        self.__depre_mod_name__ = mod_name
        self.__deprecations__ = obj_pairs

    def get(self, name: str) -> Any:
        if name not in self.__deprecations__:
            raise AttributeError(
                f"module {self.__depre_mod_name__!r} has no attribute {name!r}"
            )
        location, varname, ver = self.__deprecations__[name]
        deprecate_warn(
            f"{self.__depre_mod_name__}.{name} 现以弃用，"
            f"将于 {ver} 版本移除，使用 {location}.{varname} 代替",
            stacklevel=4,
        )
        return getattr(importlib.import_module(location), varname)

    @staticmethod
    def merge(name: str, *loaders: DeprecatedLoader) -> DeprecatedLoader:
        dic: dict[str, tuple[str, str, str]] = {}
        for loader in loaders:
            dic |= loader.__deprecations__
        return DeprecatedLoader(name, dic)


class RWContext:
    """异步读写上下文

    提供异步安全的读写上下文。在读取时可以多读，同时读写互斥。

    使用方法：

    .. code:: python

       rwc = RWContext()
       # 读时使用此控制器的安全读上下文：
       async with rwc.read():
           ...
       # 写时使用此控制器的安全写上下文：
       async with rwc.write():
           ...
    """

    def __init__(self, read_limit: int | None = None) -> None:
        """初始化异步读写上下文

        :param read_limit: 读取的数量限制，为空则不限制
        """
        self.write_semaphore = asyncio.Semaphore(1)
        self.read_semaphore = asyncio.Semaphore(read_limit) if read_limit else None
        self.read_num = 0
        self.read_num_lock = asyncio.Lock()

    @asynccontextmanager
    async def read(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步读上下文"""
        if self.read_semaphore:
            await self.read_semaphore.acquire()

        async with self.read_num_lock:
            if self.read_num == 0:
                await self.write_semaphore.acquire()
            self.read_num += 1

        try:
            yield
        finally:
            async with self.read_num_lock:
                self.read_num -= 1
                if self.read_num == 0:
                    self.write_semaphore.release()
                if self.read_semaphore:
                    self.read_semaphore.release()

    @asynccontextmanager
    async def write(self) -> AsyncGenerator[None, None]:
        """上下文管理器，展开一个关于该对象的安全异步写上下文"""
        await self.write_semaphore.acquire()
        try:
            yield
        finally:
            self.write_semaphore.release()


class SnowFlakeIdWorker:
    def __init__(self, datacenter_id: int, worker_id: int, sequence: int = 0) -> None:
        self.max_worker_id = -1 ^ (-1 << 3)
        self.max_datacenter_id = -1 ^ (-1 << 5)
        self.worker_id_shift = 12
        self.datacenter_id_shift = 12 + 3
        self.timestamp_left_shift = 12 + 3 + 5
        self.sequence_mask = -1 ^ (-1 << 12)
        self.startepoch = int(datetime(2022, 12, 11, 12, 8, 45).timestamp() * 1000)

        if worker_id > self.max_worker_id or worker_id < 0:
            raise ValueError("worker_id 值越界")
        if datacenter_id > self.max_datacenter_id or datacenter_id < 0:
            raise ValueError("datacenter_id 值越界")
        self.worker_id = worker_id
        self.datacenter_id = datacenter_id
        self.sequence = sequence
        self.last_timestamp = -1

    def _gen_timestamp(self) -> int:
        return int(time.time() * 1000)

    def get_id(self) -> int:
        timestamp = self._gen_timestamp()

        if timestamp < self.last_timestamp:
            raise ValueError(f"时钟回拨，{self.last_timestamp} 前拒绝 id 生成请求")
        if timestamp == self.last_timestamp:
            self.sequence = (self.sequence + 1) & self.sequence_mask
            if self.sequence == 0:
                timestamp = self._until_next_millis(self.last_timestamp)
        else:
            self.sequence = 0
        self.last_timestamp = timestamp
        new_id = (
            ((timestamp - self.startepoch) << self.timestamp_left_shift)
            | (self.datacenter_id << self.datacenter_id_shift)
            | (self.worker_id << self.worker_id_shift)
            | self.sequence
        )
        return new_id

    def get_b64_id(self, trim_pad: bool = True) -> str:
        id = base64.urlsafe_b64encode(
            self.get_id().to_bytes(8, byteorder="little")
        ).decode()
        if trim_pad:
            id = id.rstrip("=")
        return id

    def _until_next_millis(self, last_time: int) -> int:
        timestamp = self._gen_timestamp()
        while timestamp <= last_time:
            timestamp = self._gen_timestamp()
        return timestamp


_DEFAULT_ID_WORKER = SnowFlakeIdWorker(1, 1, 0)


def get_id() -> str:
    """从 melobot 内部 id 获取器获得一个 id 值，不保证线程安全。算法使用雪花算法

    :return: id 值
    """
    return _DEFAULT_ID_WORKER.get_b64_id()



# pylint: disable=duplicate-code
import asyncio
import json
import time
from asyncio import Future, Lock
from itertools import count

import websockets
from websockets.asyncio.client import ClientConnection
from websockets.exceptions import ConnectionClosed

from melobot.exceptions import SourceError
from melobot.io import SourceLifeSpan
from melobot.log import LogLevel

from .base import BaseIOSource
from .packet import EchoPacket, InPacket, OutPacket


class ForwardWebSocketIO(BaseIOSource):
    def __init__(
        self,
        url: str,
        max_retry: int = -1,
        retry_delay: float = 4.0,
        cd_time: float = 0,
        access_token: str | None = None,
    ) -> None:
        super().__init__(cd_time)
        self.url = url
        self.conn: ClientConnection
        self.access_token = access_token
        self.max_retry: int = max_retry
        self.retry_delay: float = retry_delay if retry_delay > 0 else 0.5

        self._in_buf: asyncio.Queue[InPacket] = asyncio.Queue()
        self._out_buf: asyncio.Queue[OutPacket] = asyncio.Queue()
        self._pre_send_time = time.time_ns()
        self._echo_table: dict[str, tuple[str, Future[EchoPacket]]] = {}

        self._tasks: list[asyncio.Task] = []
        self._opened = asyncio.Event()
        self._lock = Lock()
        self._restart_flag = asyncio.Event()

    async def _input_loop(self) -> None:
        # pylint: disable=duplicate-code
        while True:
            try:
                await self._opened.wait()

                raw_str = await self.conn.recv()
                self.logger.generic_obj(
                    "收到上报，未格式化的字符串", raw_str, level=LogLevel.DEBUG
                )
                if raw_str == "":
                    continue
                raw = json.loads(raw_str)

                if "post_type" in raw:
                    await self._in_buf.put(InPacket(time=raw["time"], data=raw))
                    continue

                echo_id = raw.get("echo")
                if echo_id in (None, ""):
                    continue

                action_type, fut = self._echo_table.pop(echo_id)
                fut.set_result(
                    EchoPacket(
                        time=int(time.time()),
                        data=raw,
                        ok=raw["status"] == "ok",
                        status=raw["retcode"],
                        action_type=action_type,
                    )
                )

            except asyncio.CancelledError:
                break

            except ConnectionClosed:
                if self.opened():
                    self._restart_flag.set()
                    asyncio.create_task(self.close())
                break

            except Exception:
                self.logger.exception("OneBot v11 正向 WebSocket IO 源输入异常")
                self.logger.generic_obj("异常点局部变量", locals(), level=LogLevel.ERROR)

    async def _output_loop(self) -> None:
        while True:
            try:
                await self._opened.wait()

                out_packet = await self._out_buf.get()
                wait_time = self.cd_time - ((time.time_ns() - self._pre_send_time) / 1e9)
                await asyncio.sleep(wait_time)
                await self.conn.send(out_packet.data)
                self._pre_send_time = time.time_ns()

            except asyncio.CancelledError:
                break

            except Exception:
                self.logger.exception("OneBot v11 正向 WebSocket IO 源输出异常")
                self.logger.generic_obj("异常点局部变量", locals(), level=LogLevel.ERROR)

    async def open(self) -> None:
        if self.opened():
            return

        async with self._lock:
            if self.opened():
                return

            headers: dict | None = None
            if self.access_token is not None:
                headers = {"Authorization": f"Bearer {self.access_token}"}

            retry_iter = count(0) if self.max_retry < 0 else range(self.max_retry + 1)
            for _ in retry_iter:
                try:
                    self.conn = await websockets.connect(
                        self.url, additional_headers=headers
                    )
                    break

                except asyncio.CancelledError:
                    raise

                except BaseException as e:
                    self.logger.warning(
                        f"连接建立失败，{self.retry_delay}s 后自动重试。错误：{e}"
                    )
                    if "403" in str(e):
                        self.logger.warning("403 错误可能是 access_token 未配置或无效")

                await asyncio.sleep(self.retry_delay)

            else:
                raise SourceError(
                    "OneBot v11 正向 WebSocket IO 源重试已达最大次数，已放弃建立连接"
                )

            self._tasks.append(asyncio.create_task(self._input_loop()))
            self._tasks.append(asyncio.create_task(self._output_loop()))
            self._opened.set()
            self.logger.info("OneBot v11 正向 WebSocket IO 源与实现端建立了连接")

            if self._restart_flag.is_set():
                self._restart_flag.clear()
                await self._hook_bus.emit(SourceLifeSpan.RESTARTED, False)

    def opened(self) -> bool:
        return self._opened.is_set()

    async def close(self) -> None:
        if not self.opened():
            return

        async with self._lock:
            if not self.opened():
                return

            self.conn.close_timeout = 2
            self._opened.clear()
            await self.conn.close()
            await self.conn.wait_closed()

            for t in self._tasks:
                t.cancel()
            if len(self._tasks):
                await asyncio.wait(self._tasks)
            self._tasks.clear()
            self.logger.info("OneBot v11 正向 WebSocket IO 源已断开连接")

            if self._restart_flag.is_set():
                asyncio.create_task(self.open())

    async def input(self) -> InPacket:
        return await self._in_buf.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        await self._out_buf.put(packet)
        if packet.echo_id is None:
            return EchoPacket(noecho=True)

        fut: Future[EchoPacket] = asyncio.get_running_loop().create_future()
        self._echo_table[packet.echo_id] = (packet.action_type, fut)
        return await fut



import asyncio
import time
from asyncio import Task
from enum import Enum

from typing_extensions import Any, Generic, TypeVar

from .ctx import LoggerCtx
from .di import inject_deps
from .log.base import LogLevel
from .typ.base import AsyncCallable, SyncOrAsyncCallable
from .utils import to_async, to_sync

HookEnumT = TypeVar("HookEnumT", bound=Enum)


class HookRunner(Generic[HookEnumT]):
    def __init__(
        self, type: HookEnumT, func: AsyncCallable[..., None], once: bool = False
    ) -> None:
        self.type = type
        self.callback: AsyncCallable[..., None] = inject_deps(func, manual_arg=True)
        self.once = once
        self._valid = True
        self._lock = asyncio.Lock()

    async def _run(self, *args: Any, **kwargs: Any) -> None:
        try:
            await self.callback(*args, **kwargs)
        except Exception:
            logger = LoggerCtx().get()
            logger.exception(f"{self.type} 类型的 hook 方法 {self.callback} 发生异常")
            logger.generic_obj(
                "异常点局部变量：",
                {"hook_runner": self, "args": args, "kwargs": kwargs},
                level=LogLevel.ERROR,
            )

    async def run(self, *args: Any, **kwargs: Any) -> None:
        if not self._valid:
            return
        if not self.once:
            return await self._run(*args, **kwargs)

        async with self._lock:
            if not self._valid:
                return
            await self._run(*args, **kwargs)
            self._valid = False


class HookBus(Generic[HookEnumT]):
    def __init__(self, type: type[HookEnumT], tag: str | None = None) -> None:
        self._hooks: dict[HookEnumT, list[HookRunner]] = {t: [] for t in list(type)}
        self._stamps: dict[HookEnumT, float] = {}
        self._tag = tag

    def set_tag(self, tag: str | None) -> None:
        self._tag = tag

    def register(
        self,
        hook_type: HookEnumT,
        hook_func: SyncOrAsyncCallable[..., None],
        once: bool = True,
    ) -> None:
        runner = HookRunner(hook_type, to_async(hook_func), once)
        self._hooks[hook_type].append(runner)

    def get_evoke_time(self, hook_type: HookEnumT) -> float:
        return self._stamps.get(hook_type, -1)

    async def emit(
        self,
        hook_type: HookEnumT,
        wait: bool = False,
        /,
        *,
        args: tuple | None = None,
        kwargs: dict[str, Any] | None = None,
        callback: SyncOrAsyncCallable[[Task[None] | None], None] | None = None,
    ) -> None:
        self._stamps[hook_type] = time.time_ns() / 1e9
        args = args if args is not None else ()
        kwargs = kwargs if kwargs is not None else {}
        logger = LoggerCtx().get()

        msg = f"<{hook_type}>（{wait = }）"  # noqa: E251, E202
        if self._tag:
            msg = f"开始 {self._tag} 的 hook: {msg}"
        else:
            msg = f"开始 hook: {msg}"
        logger.debug(msg)

        tasks = tuple(
            asyncio.create_task(runner.run(*args, **kwargs))
            for runner in self._hooks[hook_type]
        )

        if callback is not None:
            if len(tasks):
                for t in tasks:
                    t.add_done_callback(to_sync(callback))
            else:
                to_sync(callback)(None)
                return

        if wait and len(tasks):
            await asyncio.wait(tasks)



# pylint: disable=duplicate-code
import asyncio
import http
import json
import time
from asyncio import Future

import websockets
import websockets.server
from websockets import ConnectionClosed
from websockets.asyncio.server import Server, ServerConnection
from websockets.http11 import Request, Response

from melobot.io import SourceLifeSpan
from melobot.log import LogLevel

from .base import BaseIOSource
from .packet import EchoPacket, InPacket, OutPacket


class ReverseWebSocketIO(BaseIOSource):
    def __init__(
        self, host: str, port: int, cd_time: float = 0, access_token: str | None = None
    ) -> None:
        super().__init__(cd_time)
        self.host = host
        self.port = port
        self.conn: ServerConnection
        self.server: Server
        self.access_token = access_token

        self._in_buf: asyncio.Queue[InPacket] = asyncio.Queue()
        self._out_buf: asyncio.Queue[OutPacket] = asyncio.Queue()
        self._echo_table: dict[str, tuple[str, Future[EchoPacket]]] = {}
        self._pre_send_time = time.time_ns()

        self._tasks: list[asyncio.Task] = []
        self._opened = asyncio.Event()
        self._conn_requested = False
        self._req_lock = asyncio.Lock()
        self._lock = asyncio.Lock()
        self._restart_flag = asyncio.Event()

    async def _req_check(self, conn: ServerConnection, req: Request) -> Response | None:
        _headers = dict(req.headers)
        reconn_refused = "Already accepted the unique connection\n"
        auth_failed = "Authorization failed\n"

        if self._conn_requested:
            return conn.respond(http.HTTPStatus.FORBIDDEN, reconn_refused)

        async with self._req_lock:
            if self._conn_requested:
                return conn.respond(http.HTTPStatus.FORBIDDEN, reconn_refused)

            if (
                self.access_token is not None
                and _headers.get("authorization") != f"Bearer {self.access_token}"
                and _headers.get("Authorization") != f"Bearer {self.access_token}"
            ):
                self.logger.warning("OneBot 实现端的 access_token 不匹配，拒绝连接")
                return conn.respond(http.HTTPStatus.FORBIDDEN, auth_failed)

            self._conn_requested = True
            return None

    async def _input_loop(self, ws: ServerConnection) -> None:
        # pylint: disable=duplicate-code
        self.conn = ws
        self._opened.set()
        self.logger.info("实现端与 OneBot v11 反向 WebSocket IO 源建立了连接")

        if self._restart_flag.is_set():
            self._restart_flag.clear()
            await self._hook_bus.emit(SourceLifeSpan.RESTARTED, False)

        while True:
            try:
                raw_str = await self.conn.recv()
                self.logger.generic_obj(
                    "收到上报，未格式化的字符串", raw_str, level=LogLevel.DEBUG
                )
                if raw_str == "":
                    continue
                raw = json.loads(raw_str)

                if "post_type" in raw:
                    await self._in_buf.put(InPacket(time=raw["time"], data=raw))
                    continue

                echo_id = raw.get("echo")
                if echo_id in (None, ""):
                    continue

                action_type, fut = self._echo_table.pop(echo_id)
                fut.set_result(
                    EchoPacket(
                        time=int(time.time()),
                        data=raw,
                        ok=raw["status"] == "ok",
                        status=raw["retcode"],
                        action_type=action_type,
                    )
                )

            except asyncio.CancelledError:
                break

            except ConnectionClosed:
                self.logger.info(
                    "实现端与 OneBot v11 反向 WebSocket IO 源已断连，等待连接中"
                )
                self._opened.clear()
                self._restart_flag.set()
                self._conn_requested = False
                break

            except Exception:
                self.logger.exception("OneBot v11 反向 WebSocket IO 源输入异常")
                self.logger.generic_obj("异常点局部变量", locals(), level=LogLevel.ERROR)

    async def _output_loop(self) -> None:
        while True:
            try:
                await self._opened.wait()

                out_packet = await self._out_buf.get()
                wait_time = self.cd_time - ((time.time_ns() - self._pre_send_time) / 1e9)
                await asyncio.sleep(wait_time)
                await self.conn.send(out_packet.data)
                self._pre_send_time = time.time_ns()

            except asyncio.CancelledError:
                break

            except Exception:
                self.logger.exception("OneBot v11 反向 WebSocket IO 源输出异常")
                self.logger.generic_obj("异常点局部变量", locals(), level=LogLevel.ERROR)

    async def open(self) -> None:
        if self.opened():
            return

        async with self._lock:
            if self.opened():
                return

            self.server = await websockets.serve(
                self._input_loop, self.host, self.port, process_request=self._req_check
            )
            self._tasks.append(asyncio.create_task(self._output_loop()))
            self.logger.info("OneBot v11 反向 WebSocket IO 源启动了服务，等待连接中")
            await self._opened.wait()

    def opened(self) -> bool:
        return self._opened.is_set()

    async def close(self) -> None:
        if not self.opened():
            return

        async with self._lock:
            if not self.opened():
                return

            self._opened.clear()
            self.server.close()  # type: ignore[func-returns-value]
            await self.server.wait_closed()

            for t in self._tasks:
                t.cancel()
            if len(self._tasks):
                await asyncio.wait(self._tasks)
            self._tasks.clear()
            self.logger.info("OneBot v11 反向 WebSocket IO 源的服务已停止运行")

    async def input(self) -> InPacket:
        return await self._in_buf.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        await self._out_buf.put(packet)
        if packet.echo_id is None:
            return EchoPacket(noecho=True)

        fut: Future[EchoPacket] = asyncio.get_running_loop().create_future()
        self._echo_table[packet.echo_id] = (packet.action_type, fut)
        return await fut



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from typing_extensions import Callable, final, overload

from ..exceptions import PluginLoadError
from ..handle.base import Flow
from ..mixin import HookMixin
from ..typ.base import P, T
from .ipc import AsyncShare, SyncShare


class PluginLifeSpan(Enum):
    """插件生命周期的枚举"""

    INITED = "i"


@dataclass(frozen=True)
class PluginInfo:
    """插件信息类，用于添加描述信息"""

    version: str = ""
    desc: str = ""
    docs: Path | None = None
    keywords: list[str] | None = None
    url: str = ""
    author: str = ""


class PluginPlanner(HookMixin[PluginLifeSpan]):
    """插件管理器类

    用于声明一个插件，并为插件添加功能
    """

    def __init__(
        self,
        version: str,
        flows: list[Flow] | None = None,
        shares: list[SyncShare | AsyncShare] | None = None,
        funcs: list[Callable] | None = None,
        *,
        info: PluginInfo | None = None,
    ) -> None:
        """插件管理器初始化

        :param version: 版本号
        :param flows: 事件流列表。可以先指定为空，后续使用 :meth:`use` 绑定
        :param shares: 共享对象列表。可以先指定为空，后续使用 :meth:`use` 绑定
        :param funcs: 导出函数列表。可以先指定为空，后续使用 :meth:`use` 绑定
        :param info: 插件信息
        """
        super().__init__(hook_type=PluginLifeSpan)
        self.version = version
        self.init_flows = [] if flows is None else flows
        self.shares = [] if shares is None else shares
        self.funcs = [] if funcs is None else funcs
        self.info = PluginInfo() if info is None else info

        self._pname: str = ""
        self._built: bool = False
        self._plugin: Plugin

    @final
    def __p_build__(self, name: str) -> Plugin:
        if not self._built:
            self._pname = name
            self._hook_bus.set_tag(name)
            self._plugin = Plugin(self)
            self._built = True
        return self._plugin

    @overload
    def use(self, obj: Flow) -> Flow: ...
    @overload
    def use(self, obj: SyncShare[T]) -> SyncShare[T]: ...  # type: ignore[overload-overlap]
    @overload
    def use(self, obj: AsyncShare[T]) -> AsyncShare[T]: ...  # type: ignore[overload-overlap]
    @overload
    def use(self, obj: Callable[P, T]) -> Callable[P, T]: ...

    @final
    def use(self, obj: T) -> T:
        """装饰器

        绑定一个组件（流，共享对象，导出函数），标记插件创建后使用该组件。

        :param obj: 可用的组件
        :return: 被绑定的组件本身
        """
        if isinstance(obj, Flow):
            self.init_flows.append(obj)
        elif isinstance(obj, (SyncShare, AsyncShare)):
            self.shares.append(obj)
        elif callable(obj):
            self.funcs.append(obj)
        else:
            raise PluginLoadError(f"插件无法使用 {type(obj)} 类型的对象")
        return obj


class Plugin:
    def __init__(self, planner: PluginPlanner) -> None:
        self.planner = planner
        self.name = planner._pname
        self.hook_bus = planner._hook_bus

        self.shares = planner.shares
        self.funcs = planner.funcs
        self.init_flows = planner.init_flows



import time

from melobot.protocols.onebot.v11.io import packet
from tests.base import *


async def test_packet():
    p = packet.InPacket(time=time.time(), data={"key": "value"})
    p = packet.OutPacket(
        time=time.time(),
        action_type="action",
        action_params={"key": "value"},
        echo_id="123456",
    )
    p = packet.EchoPacket(time=time.time(), data={"key": "value"})



from melobot.handle.base import Flow, FlowNode
from tests.base import *


def node():
    async def node():
        return True

    n = FlowNode(node)
    n.name = str(id(n))
    return n


async def test_flow_link():
    a = Flow("a", [node()], [node(), node()], [node()])
    b = Flow("b", [node(), node()])
    c = Flow("c", [node()])
    nf = a.link(b)
    assert sum(len(info.nexts) for info in nf.graph.values()) == 5
    assert len(nf.starts) == 3
    assert len(nf.ends) == 1
    nf = a.link(Flow("test"))
    assert sum(len(info.nexts) for info in nf.graph.values()) == 1
    assert len(nf.starts) == 3
    assert len(nf.ends) == 3
    nf = Flow("test").link(b)
    assert sum(len(info.nexts) for info in nf.graph.values()) == 1
    assert len(nf.starts) == 1
    assert len(nf.ends) == 1
    nf = Flow("test").link(c)
    assert sum(len(info.nexts) for info in nf.graph.values()) == 0
    assert len(nf.starts) == 1
    assert len(nf.ends) == 1



from argparse import Namespace
from pathlib import Path

from melobot.plugin.load import PluginInitHelper


def main(args: Namespace) -> None:
    if not len(args.files):
        print("未提供插件目录参数")
        return

    p_dirs = set(Path(p_dir) for p_dir in args.files)
    PluginInitHelper.run_init(*p_dirs, load_depth=args.depth)
    print(f"已完成 {len(args.files)} 个插件的初始化")



from __future__ import annotations

import asyncio
from os import PathLike

from typing_extensions import Any, Callable, Iterable, Literal, Optional, Sequence, cast

from melobot.adapter import (
    AbstractEchoFactory,
    AbstractEventFactory,
    AbstractOutputFactory,
    ActionHandle,
)
from melobot.adapter import Adapter as RootAdapter
from melobot.adapter import Content, EchoT
from melobot.adapter import content as mc
from melobot.ctx import Context
from melobot.exceptions import AdapterError
from melobot.handle import try_get_event
from melobot.typ import AsyncCallable, SyncOrAsyncCallable
from melobot.utils import to_async, to_coro

from ..const import ACTION_TYPE_KEY_NAME, PROTOCOL_IDENTIFIER, P, T
from ..io.base import BaseIOSource
from ..io.packet import EchoPacket, InPacket, OutPacket
from . import action as ac
from . import echo as ec
from . import event as ev
from . import segment as se

_ValidateHandler = AsyncCallable[[dict[str, Any], Exception], None]


class ValidateHandleMixin:
    def __init__(self) -> None:
        self.validate_handlers: list[_ValidateHandler] = []

    def add_validate_handler(self, callback: _ValidateHandler) -> None:
        self.validate_handlers.append(callback)

    async def validate_handle(
        self, data: dict[str, Any], func: Callable[[dict[str, Any]], T]
    ) -> T:
        try:
            return func(data)
        except Exception as e:
            tasks = map(
                asyncio.create_task,
                map(
                    to_coro,
                    (cb(data, e) for cb in self.validate_handlers),
                ),
            )
            if len(self.validate_handlers):
                await asyncio.wait(tasks)
        return func(data)


class EventFactory(AbstractEventFactory[InPacket, ev.Event], ValidateHandleMixin):
    async def create(self, packet: InPacket) -> ev.Event:
        return await self.validate_handle(packet.data, ev.Event.resolve)


class OutputFactory(AbstractOutputFactory[OutPacket, ac.Action]):
    async def create(self, action: ac.Action) -> OutPacket:
        return OutPacket(
            data=action.flatten(),
            action_type=action.type,
            action_params=action.params,
            echo_id=action.id if action.need_echo else None,
        )


class EchoFactory(AbstractEchoFactory[EchoPacket, ec.Echo], ValidateHandleMixin):
    async def create(self, packet: EchoPacket) -> ec.Echo | None:
        if packet.noecho:
            return None

        data = packet.data
        data[ACTION_TYPE_KEY_NAME] = packet.action_type
        return await self.validate_handle(data, ec.Echo.resolve)


class EchoRequireCtx(Context[bool]):
    def __init__(self) -> None:
        super().__init__("ONEBOT_V11_ECHO_REQUIRE", LookupError)


class Adapter(
    RootAdapter[
        EventFactory, OutputFactory, EchoFactory, ac.Action, BaseIOSource, BaseIOSource
    ]
):
    def __init__(self) -> None:
        super().__init__(
            PROTOCOL_IDENTIFIER, EventFactory(), OutputFactory(), EchoFactory()
        )

    def when_validate_error(self, validate_type: Literal["event", "echo"]) -> Callable[
        [SyncOrAsyncCallable[[dict[str, Any], Exception], None]],
        AsyncCallable[[dict[str, Any], Exception], None],
    ]:
        def when_validate_error_wrapper(
            func: SyncOrAsyncCallable[[dict[str, Any], Exception], None]
        ) -> AsyncCallable[[dict[str, Any], Exception], None]:
            f = to_async(func)
            if validate_type == "event":
                self._event_factory.add_validate_handler(f)
            elif validate_type == "echo":
                self._echo_factory.add_validate_handler(f)
            else:
                raise AdapterError("无效的验证类型，合法值是 'event', 'echo' 之一")
            return f

        return when_validate_error_wrapper

    async def call_output(self, action: ac.Action) -> tuple[ActionHandle, ...]:
        """输出行为的底层方法

        :param action: 行为对象
        :return: :class:`.ActionHandle` 元组
        """
        if EchoRequireCtx().try_get():
            action.need_echo = True
        return await super().call_output(action)

    def with_echo(
        self, func: AsyncCallable[P, tuple[ActionHandle[EchoT | None], ...]]
    ) -> AsyncCallable[P, tuple[ActionHandle[EchoT], ...]]:
        async def with_echo_wrapped(
            *args: P.args, **kwargs: P.kwargs
        ) -> tuple[ActionHandle[EchoT], ...]:
            with EchoRequireCtx().unfold(True):
                handles = await func(*args, **kwargs)
            return cast(tuple[ActionHandle[EchoT], ...], handles)

        return with_echo_wrapped

    async def __send_text__(
        self, text: str
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(text)

    async def __send_media__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.MediaContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_image__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.ImageContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_audio__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.AudioContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_voice__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.VoiceContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_video__(
        self,
        name: str,
        raw: bytes | None = None,
        url: str | None = None,
        mimetype: str | None = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs(
                [mc.VideoContent(name=name, url=url, raw=raw, mimetype=mimetype)]
            )[0]
        )

    async def __send_file__(
        self, name: str, path: str | PathLike[str]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(
            se.contents_to_segs([mc.FileContent(name=name, flag=str(path))])[0]
        )

    async def __send_refer__(
        self, event: ev.RootEvent, contents: Sequence[Content] | None = None
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"提供的事件不是 {ev.MessageEvent.__qualname__} 类型，无法用于发送 refer 消息"
            )

        segs = se.contents_to_segs(list(contents)) if contents else []
        segs.insert(0, se.ReplySegment(str(event.message_id)))
        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_custom(segs, group_id=event.group_id)
        return await self.send_custom(segs, user_id=event.user_id)

    async def __send_resource__(
        self, name: str, url: str
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.send(se.contents_to_segs([mc.ResourceContent(name, url)])[0])

    async def send(
        self, msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_custom(msgs, group_id=event.group_id)
        return await self.send_custom(msgs, user_id=event.user_id)

    async def send_reply(
        self, msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict]
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        kwargs: dict[str, Any] = {
            "msgs": ac.msgs_to_dicts(se.ReplySegment(str(event.message_id)))
            + ac.msgs_to_dicts(msgs)
        }
        if isinstance(event, ev.GroupMessageEvent):
            kwargs["group_id"] = event.group_id
        else:
            kwargs["user_id"] = event.user_id
        return await self.call_output(ac.SendMsgAction(**kwargs))

    async def send_custom(
        self,
        msgs: str | se.Segment | Iterable[se.Segment] | dict | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> tuple[ActionHandle[ec.SendMsgEcho | None], ...]:
        return await self.call_output(ac.SendMsgAction(msgs, user_id, group_id))

    async def send_forward(
        self, msgs: Iterable[se.NodeSegment]
    ) -> tuple[ActionHandle[ec.SendForwardMsgEcho | None], ...]:
        event = try_get_event()
        if not isinstance(event, ev.MessageEvent):
            raise AdapterError(
                f"当前上下文中不存在事件，或事件不为 {ev.MessageEvent.__qualname__} 类型，无法发送消息"
            )

        if isinstance(event, ev.GroupMessageEvent):
            return await self.send_forward_custom(msgs, group_id=event.group_id)
        return await self.send_forward_custom(msgs, user_id=event.user_id)

    async def send_forward_custom(
        self,
        msgs: Iterable[se.NodeSegment] | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> tuple[ActionHandle[ec.SendForwardMsgEcho | None], ...]:
        return await self.call_output(ac.SendForwardMsgAction(msgs, user_id, group_id))

    async def delete_msg(
        self, msg_id: int
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.DeleteMsgAction(msg_id))

    async def get_msg(
        self, msg_id: int
    ) -> tuple[ActionHandle[ec.GetMsgEcho | None], ...]:
        return await self.call_output(ac.GetMsgAction(msg_id))

    async def get_forward_msg(
        self, forward_id: str
    ) -> tuple[ActionHandle[ec.GetForwardMsgEcho | None], ...]:
        return await self.call_output(ac.GetForwardMsgAction(forward_id))

    async def send_like(
        self, user_id: int, times: int = 1
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SendLikeAction(user_id, times))

    async def set_group_kick(
        self, group_id: int, user_id: int, later_reject: bool = False
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupKickAction(group_id, user_id, later_reject)
        )

    async def set_group_ban(
        self, group_id: int, user_id: int, duration: int = 30 * 60
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupBanAction(group_id, user_id, duration))

    async def set_group_anonymous_ban(
        self,
        group_id: int,
        anonymous: ac.SetGroupAnonymousBanAction.AnonymousDict,
        anonymous_flag: str,
        duration: int = 30 * 60,
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupAnonymousBanAction(group_id, anonymous, anonymous_flag, duration)
        )

    async def set_group_whole_ban(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupWholeBanAction(group_id, enable))

    async def set_group_admin(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupAdminAction(group_id, enable))

    async def set_group_anonymous(
        self, group_id: int, enable: bool = True
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupAnonymousAction(group_id, enable))

    async def set_group_card(
        self, group_id: int, user_id: int, card: str = ""
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupCardAction(group_id, user_id, card))

    async def set_group_name(
        self, group_id: int, name: str
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupNameAction(group_id, name))

    async def set_group_leave(
        self, group_id: int, is_dismiss: bool = False
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetGroupLeaveAction(group_id, is_dismiss))

    async def set_group_special_title(
        self, group_id: int, user_id: int, title: str = "", duration: int = -1
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupSpecialTitleAction(group_id, user_id, title, duration)
        )

    async def set_friend_add_request(
        self, add_flag: str, approve: bool = True, remark: str = ""
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetFriendAddRequestAction(add_flag, approve, remark)
        )

    async def set_group_add_request(
        self,
        add_flag: str,
        add_type: Literal["add", "invite"],
        approve: bool = True,
        reason: str = "",
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(
            ac.SetGroupAddRequestAction(add_flag, add_type, approve, reason)
        )

    async def get_login_info(self) -> tuple[ActionHandle[ec.GetLoginInfoEcho], ...]:
        return await self.call_output(ac.GetLoginInfoAction())

    async def get_stranger_info(
        self, user_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetStrangerInfoEcho], ...]:
        return await self.call_output(ac.GetStrangerInfoAction(user_id, no_cache))

    async def get_friend_list(self) -> tuple[ActionHandle[ec.GetFriendListEcho], ...]:
        return await self.call_output(ac.GetFriendlistAction())

    async def get_group_info(
        self, group_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetGroupInfoEcho], ...]:
        return await self.call_output(ac.GetGroupInfoAction(group_id, no_cache))

    async def get_group_list(self) -> tuple[ActionHandle[ec.GetGroupListEcho], ...]:
        return await self.call_output(ac.GetGrouplistAction())

    async def get_group_member_info(
        self, group_id: int, user_id: int, no_cache: bool = False
    ) -> tuple[ActionHandle[ec.GetGroupMemberInfoEcho], ...]:
        return await self.call_output(
            ac.GetGroupMemberInfoAction(group_id, user_id, no_cache)
        )

    async def get_group_member_list(
        self, group_id: int
    ) -> tuple[ActionHandle[ec.GetGroupMemberListEcho], ...]:
        return await self.call_output(ac.GetGroupMemberlistAction(group_id))

    async def get_group_honor_info(
        self,
        group_id: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> tuple[ActionHandle[ec.GetGroupHonorInfoEcho], ...]:
        return await self.call_output(ac.GetGroupHonorInfoAction(group_id, type))

    async def get_cookies(
        self, domain: str = ""
    ) -> tuple[ActionHandle[ec.GetCookiesEcho], ...]:
        return await self.call_output(ac.GetCookiesAction(domain))

    async def get_csrf_token(self) -> tuple[ActionHandle[ec.GetCsrfTokenEcho], ...]:
        return await self.call_output(ac.GetCsrfTokenAction())

    async def get_credentials(
        self, domain: str = ""
    ) -> tuple[ActionHandle[ec.GetCredentialsEcho], ...]:
        return await self.call_output(ac.GetCredentialsAction(domain))

    async def get_record(
        self, file: str, out_format: str
    ) -> tuple[ActionHandle[ec.GetRecordEcho], ...]:
        return await self.call_output(ac.GetRecordAction(file, out_format))

    async def get_image(self, file: str) -> tuple[ActionHandle[ec.GetImageEcho], ...]:
        return await self.call_output(ac.GetImageAction(file))

    async def can_send_image(self) -> tuple[ActionHandle[ec.CanSendImageEcho], ...]:
        return await self.call_output(ac.CanSendImageAction())

    async def can_send_record(self) -> tuple[ActionHandle[ec.CanSendRecordEcho], ...]:
        return await self.call_output(ac.CanSendRecordAction())

    async def get_status(self) -> tuple[ActionHandle[ec.GetStatusEcho], ...]:
        return await self.call_output(ac.GetStatusAction())

    async def get_version_info(self) -> tuple[ActionHandle[ec.GetVersionInfoEcho], ...]:
        return await self.call_output(ac.GetVersionInfoAction())

    async def set_restart(
        self, delay: int = 0
    ) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.SetRestartAction(delay))

    async def clean_cache(self) -> tuple[ActionHandle[ec.EmptyEcho | None], ...]:
        return await self.call_output(ac.CleanCacheAction())



from os import PathLike

from typing_extensions import Sequence

from ..ctx import EventOrigin, FlowCtx
from .base import Adapter
from .content import Content
from .model import ActionHandle, Event

_CTX = FlowCtx()


async def send_text(text: str) -> tuple[ActionHandle, ...]:
    """通用文本输出方法"""
    return await _get_ctx_adapter().__send_text__(text)


async def send_media(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用媒体内容输出方法"""
    return await _get_ctx_adapter().__send_media__(name, raw, url, mimetype)


async def send_image(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用图像内容输出方法"""
    return await _get_ctx_adapter().__send_image__(name, raw, url, mimetype)


async def send_audio(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用音频内容输出方法"""
    return await _get_ctx_adapter().__send_audio__(name, raw, url, mimetype)


async def send_voice(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用语音内容输出方法"""
    return await _get_ctx_adapter().__send_voice__(name, raw, url, mimetype)


async def send_video(
    name: str,
    raw: bytes | None = None,
    url: str | None = None,
    mimetype: str | None = None,
) -> tuple[ActionHandle, ...]:
    """通用视频内容输出方法"""
    return await _get_ctx_adapter().__send_video__(name, raw, url, mimetype)


async def send_file(name: str, path: str | PathLike[str]) -> tuple[ActionHandle, ...]:
    """通用文件输出方法"""
    return await _get_ctx_adapter().__send_file__(name, path)


async def send_refer(
    event: Event, contents: Sequence[Content] | None = None
) -> tuple[ActionHandle, ...]:
    """通用过往事件引用输出方法"""
    return await _get_ctx_adapter().__send_refer__(event, contents)


async def send_resource(name: str, url: str) -> tuple[ActionHandle, ...]:
    """通用其他资源输出方法"""
    return await _get_ctx_adapter().__send_resource__(name, url)


def _get_ctx_adapter() -> Adapter:
    event = _CTX.get_event()
    return EventOrigin.get_origin(event).adapter



from ..adapter.base import Adapter
from ..io.base import AbstractInSource, AbstractOutSource
from ..typ.cls import BetterABC, abstractattr


class ProtocolStack(BetterABC):
    """协议栈抽象类

    子类需要把以下属性按 :func:`.abstractattr` 的要求实现
    """

    inputs: set[AbstractInSource] = abstractattr()
    """该协议栈兼容的输入源序列

       :meta hide-value:
    """

    outputs: set[AbstractOutSource] = abstractattr()
    """该协议栈兼容的输出源序列

       :meta hide-value:
    """

    adapter: Adapter = abstractattr()
    """该协议栈兼容的适配器

       :meta hide-value:
    """



from melobot.protocols import ProtocolStack
from melobot.utils.common import DeprecatedLoader as _DeprecatedLoader

from .. import __version__
from .adapter import *
from .const import (
    PROTOCOL_IDENTIFIER,
    PROTOCOL_NAME,
    PROTOCOL_SUPPORT_AUTHOR,
    PROTOCOL_VERSION,
)
from .handle import _LOADER as _HANDLE_LOADER
from .handle import on_at_qq, on_event, on_message, on_meta, on_notice, on_request
from .io import *
from .utils import _LOADER as _UTILS_LOADER
from .utils import *


class OneBotV11Protocol(ProtocolStack):
    def __init__(self, *srcs: BaseSource) -> None:
        super().__init__()
        self.adapter = Adapter()
        self.inputs = set()
        self.outputs = set()

        for src in srcs:
            if not isinstance(src, BaseSource):
                raise TypeError(f"不支持的 OneBot v11 源类型: {type(src)}")
            if isinstance(src, BaseInSource):
                self.inputs.add(src)
            if isinstance(src, BaseOutSource):
                self.outputs.add(src)


_LOADER = _DeprecatedLoader.merge(__name__, _HANDLE_LOADER, _UTILS_LOADER)


def __getattr__(name: str) -> Any:
    return _LOADER.get(name)



from __future__ import annotations

import re
from abc import abstractmethod
from functools import partial

from typing_extensions import Any, Callable, Coroutine

from melobot.exceptions import UtilValidateError
from melobot.typ import BetterABC, LogicMode


class Matcher(BetterABC):
    """匹配器基类"""

    def __init__(self) -> None:
        super().__init__()

    def __and__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilValidateError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.AND, self, other)

    def __or__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilValidateError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.OR, self, other)

    def __invert__(self) -> WrappedMatcher:
        return WrappedMatcher(LogicMode.NOT, self)

    def __xor__(self, other: Matcher) -> WrappedMatcher:
        if not isinstance(other, Matcher):
            raise UtilValidateError(
                f"联合匹配器定义时出现了非匹配器对象，其值为：{other}"
            )
        return WrappedMatcher(LogicMode.XOR, self, other)

    @abstractmethod
    async def match(self, text: str) -> bool:
        """匹配器匹配方法

        任何匹配器应该实现此抽象方法。

        :param text: 消息事件的文本内容
        :return: 是否匹配
        """
        raise NotImplementedError


class WrappedMatcher(Matcher):
    """合并匹配器

    在两个 :class:`Matcher` 对象间使用 | & ^ ~ 运算符即可返回合并匹配器
    """

    def __init__(
        self,
        mode: LogicMode,
        matcher1: Matcher,
        matcher2: Matcher | None = None,
    ) -> None:
        """初始化一个合并匹配器

        :param mode: 合并匹配的逻辑模式
        :param matcher1: 匹配器1
        :param matcher2: 匹配器2
        """
        super().__init__()
        self.mode = mode
        self.m1, self.m2 = matcher1, matcher2

    async def match(self, text: str) -> bool:
        m2_match: Callable[[], Coroutine[Any, Any, bool]] | None = (
            partial(self.m2.match, text) if self.m2 is not None else None
        )
        return await LogicMode.async_short_calc(
            self.mode, partial(self.m1.match, text), m2_match
        )


class StartMatcher(Matcher):
    """字符串起始匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串起始匹配器

        `target` 为字符串时，只进行一次起始匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行起始匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text[: len(self.target)] == self.target
        res_seq = [text[: len(s)] == s for s in self.target]
        return LogicMode.seq_calc(self.mode, res_seq)


class ContainMatcher(Matcher):
    """字符串包含匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串包含匹配器

        `target` 为字符串时，只进行一次包含匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行包含匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return self.target in text
        res_seq = [s in text for s in self.target]
        return LogicMode.seq_calc(self.mode, res_seq)


class EndMatcher(Matcher):
    """字符串结尾匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串结尾匹配器

        `target` 为字符串时，只进行一次结尾匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行结尾匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text[-len(self.target) :] == self.target
        res_seq = [text[-len(s) :] == s for s in self.target]
        return LogicMode.seq_calc(self.mode, res_seq)


class FullMatcher(Matcher):
    """字符串全匹配器"""

    def __init__(self, target: str | list[str], mode: LogicMode = LogicMode.OR) -> None:
        """初始化一个字符串全匹配器

        `target` 为字符串时，只进行一次全匹配，即判断是否匹配成功。
        `target` 为字符串列表时，所有字符串都进行全匹配，再将所有结果使用给定
        `mode` 计算是否匹配成功。

        :param target: 匹配目标
        :param mode: 匹配模式
        """
        super().__init__()
        self.target = target
        self.mode = mode

    async def match(self, text: str) -> bool:
        if isinstance(self.target, str):
            return text == self.target
        res_seq = [text == s for s in self.target]
        return LogicMode.seq_calc(self.mode, res_seq)


class RegexMatcher(Matcher):
    """字符串正则匹配器"""

    def __init__(self, regex_pattern: str, regex_flags: Any = 0) -> None:
        """初始化一个字符串正则匹配器

        :param regex_pattern: 正则 pattern
        :param regex_flags: 正则 flag，默认不使用
        """
        super().__init__()
        self.pattern = regex_pattern
        self.flag = regex_flags

    async def match(self, text: str) -> bool:
        return len(re.findall(self.pattern, text, self.flag)) > 0



import json
from asyncio import Queue, create_task

import aiohttp
import aiohttp.web

from melobot.ctx import LoggerCtx
from melobot.log.base import Logger
from melobot.protocols.onebot.v11.adapter import action
from melobot.protocols.onebot.v11.io.duplex_http import HttpIO
from melobot.protocols.onebot.v11.io.packet import OutPacket
from melobot.utils import singleton, to_async
from tests.base import *

_TEST_EVENT_DICT = {
    "time": 1725292489,
    "self_id": 123456,
    "post_type": "message",
    "message_type": "group",
    "sub_type": "normal",
    "sender": {
        "age": 0,
        "nickname": "这是一个群昵称",
        "sex": "unknown",
        "user_id": 1574260633,
        "area": "",
        "card": "",
        "level": "",
        "role": "member",
        "title": "",
    },
    "message_id": -1234567890,
    "font": 0,
    "message": "",
    "user_id": 1574260633,
    "anonymous": None,
    "group_id": 535705163,
    "raw_message": "",
}
_TEST_ACTION = action.SendMsgAction("hello world!", 123)
_TEST_ECHO_DICT = {
    "time": 1725292489,
    "status": "ok",
    "retcode": 0,
    "data": {"hello": "world!"},
}


async def put_input(s: str):
    async with aiohttp._ClientSession() as session:
        async with session.post("http://localhost:9090", json=json.loads(s)) as resp:
            return


_OUT_BUF = Queue()


@singleton
class MockClientSession:
    async def post(self, url, *args, **kwargs):
        await _OUT_BUF.put(
            json.dumps({"action": url.split("/")[-1], "params": kwargs["json"]})
        )
        resp = aiohttp.web.Response(status=200)
        resp.json = to_async(lambda: _TEST_ECHO_DICT | {"echo": _TEST_ACTION.id})
        return resp

    async def close(self):
        return


async def test_http(monkeypatch) -> None:
    with LoggerCtx().unfold(Logger()):
        aiohttp._ClientSession = aiohttp.ClientSession
        monkeypatch.setattr(aiohttp, "ClientSession", lambda: MockClientSession())
        io = HttpIO("localhost", 8080, "localhost", 9090)
        create_task(put_input(json.dumps(_TEST_EVENT_DICT)))

        async with io:
            await io.input()

            await put_input(json.dumps(_TEST_EVENT_DICT))
            pak = await io.input()
            assert pak.data == _TEST_EVENT_DICT

            pak = await io.output(
                OutPacket(
                    data=_TEST_ACTION.extract(),
                    action_type=_TEST_ACTION.extract()["action"],
                    action_params=_TEST_ACTION.extract()["params"],
                )
            )
            await _OUT_BUF.get() == _TEST_ACTION.flatten()
            assert pak.noecho

            _TEST_ACTION.set_echo(True)
            t = create_task(
                io.output(
                    OutPacket(
                        data=_TEST_ACTION.extract(),
                        action_type=_TEST_ACTION.extract()["action"],
                        action_params=_TEST_ACTION.extract()["params"],
                        echo_id=_TEST_ACTION.extract()["echo"],
                    )
                )
            )
            await _OUT_BUF.get() == _TEST_ACTION.flatten()
            pak = await t
            assert pak.data["data"]["hello"] == _TEST_ECHO_DICT["data"]["hello"]
            assert pak.ok == (_TEST_ECHO_DICT["status"] == "ok")
            assert pak.status == _TEST_ECHO_DICT["retcode"]



from __future__ import annotations

import io
import logging
import logging.config
import logging.handlers
import os
import sys
import traceback
import types
from abc import abstractmethod
from contextlib import contextmanager
from enum import Enum
from inspect import currentframe
from logging import CRITICAL, DEBUG, ERROR, INFO, WARNING
from logging import Logger as _Logger
from logging import _srcfile as _LOGGING_SRC_FILE

import colorlog
import rich.console
import rich.pretty
from better_exceptions import ExceptionFormatter
from rich.highlighter import Highlighter, ReprHighlighter
from rich.style import Style
from rich.text import Text
from typing_extensions import Any, Callable, Generator, Literal, Optional

from ..typ._enum import VoidType
from ..typ.base import T
from ..typ.cls import BetterABC
from ..utils.common import singleton

_CONSOLE_IO = io.StringIO()
_CONSOLE = rich.console.Console(file=_CONSOLE_IO, record=True, color_system="256")
_REPR_HIGHLIGHTER = ReprHighlighter()
_HIGH_LIGHTWORDS = ["GET", "POST", "HEAD", "PUT", "DELETE", "OPTIONS", "TRACE", "PATCH"]


class _StyleHighlighter(Highlighter):
    def __init__(self, style: Style | None) -> None:
        super().__init__()
        self.style = style

    def highlight(self, text: Text) -> None:
        if self.style:
            text.stylize(self.style)


def _get_rich_object(
    obj: object,
    max_len: Optional[int] = 2000,
    style: Style | None = None,
    no_color: bool = False,
) -> tuple[str, str]:
    if no_color:
        hl = _StyleHighlighter(style=None)
    elif style:
        hl = _StyleHighlighter(style)
    else:
        hl = None

    _CONSOLE.print(
        rich.pretty.Pretty(
            obj,
            highlighter=hl,
            indent_guides=True,
            max_string=max_len,
            overflow="ignore",
            expand_all=True,
        ),
        crop=False,
    )
    colored_str = _CONSOLE_IO.getvalue().rstrip("\n")
    _CONSOLE_IO.seek(0)
    _CONSOLE_IO.truncate(0)
    return colored_str, _CONSOLE.export_text().rstrip("\n")


def _get_rich_repr(
    s: str, style: Style | None = None, no_color: bool = False
) -> tuple[str, str]:
    if no_color:
        msg = Text(s)
    elif style:
        msg = Text(s, style=style)
    else:
        msg = _REPR_HIGHLIGHTER(Text(s))
        msg.highlight_words(_HIGH_LIGHTWORDS, "logging.keyword")

    _CONSOLE.print(msg)
    colored_str = _CONSOLE_IO.getvalue()[:-1]
    _CONSOLE_IO.seek(0)
    _CONSOLE_IO.truncate(0)
    return colored_str, _CONSOLE.export_text()[:-1]


_EXC_FORMATTER = ExceptionFormatter(colored=True)
_NO_COLOR_EXC_FORMATTER = ExceptionFormatter(colored=False)


class _NoErrFilter(logging.Filter):
    def filter(self, record: logging.LogRecord) -> bool:
        return logging.DEBUG <= record.levelno < logging.ERROR


class _MeloLogFilter(logging.Filter):
    def __init__(
        self,
        name: str = "",
        yellow_warn: bool = True,
        red_error: bool = True,
        legacy: bool = False,
    ) -> None:
        super().__init__(name)
        self._obj: Any = VoidType.VOID
        self._yellow_style = Style(color="yellow")
        self._red_style = Style(color="red")
        self._enable_yellow_warn = yellow_warn
        self._enable_red_error = red_error
        self._legacy = legacy

    def set_obj(self, obj: Any) -> None:
        self._obj = obj

    def clear_obj(self) -> None:
        self._obj = VoidType.VOID

    @contextmanager
    def on_obj(self, obj: Any) -> Generator[None, None, None]:
        try:
            self.set_obj(obj)
            yield
        finally:
            self.clear_obj()

    def filter(self, record: logging.LogRecord) -> Literal[True]:
        msg = str(record.msg)
        if record.args:
            msg = msg % record.args
        record.msg_str = msg

        record.mod_name, record.func_name, record.func_lineno = _current_finfo()
        self._fill_msg_and_obj(msg, record)
        return True

    def _fill_msg_and_obj(self, msg: str, record: logging.LogRecord) -> None:
        yellow_style = self._yellow_style
        red_style = self._red_style
        yellow_warn = self._enable_yellow_warn
        red_error = self._enable_red_error

        if self._legacy:
            record.legacy_msg_str, record.colored_msg_str, record.msg_str = msg, "", msg

            if self._obj is VoidType.VOID:
                record.legacy_obj, record.obj = "", ""
            else:
                record.legacy_obj = record.obj = _get_rich_object(
                    self._obj, no_color=True
                )[1]
            record.colored_obj = ""
            return

        record.legacy_msg_str = ""
        record.legacy_obj = ""

        if red_error and record.levelno >= ERROR:
            record.colored_msg_str, record.msg_str = _get_rich_repr(msg, red_style)
        elif yellow_warn and record.levelno >= WARNING:
            record.colored_msg_str, record.msg_str = _get_rich_repr(msg, yellow_style)
        else:
            record.colored_msg_str, record.msg_str = _get_rich_repr(msg)

        if self._obj is VoidType.VOID:
            record.colored_obj, record.obj = "", ""
        elif red_error and record.levelno >= ERROR:
            record.legacy_obj = record.obj = _get_rich_object(self._obj, no_color=True)[1]
            record.colored_obj = ""
        elif yellow_warn and record.levelno >= WARNING:
            record.legacy_obj = record.obj = _get_rich_object(self._obj, no_color=True)[1]
            record.colored_obj = ""
        else:
            record.colored_obj, record.obj = _get_rich_object(self._obj)


class LogLevel(int, Enum):
    """日志等级枚举"""

    CRITICAL = CRITICAL
    DEBUG = DEBUG
    ERROR = ERROR
    INFO = INFO
    WARNING = WARNING


_FILE = os.path.normcase(_get_rich_object.__code__.co_filename)


def _is_internal_frame(frame: types.FrameType) -> bool:
    filename = os.path.normcase(frame.f_code.co_filename)
    return filename in (_FILE, _LOGGING_SRC_FILE) or (
        "importlib" in filename and "_bootstrap" in filename
    )


def _current_finfo() -> tuple[str, str, int]:
    frame = currentframe()
    while frame:
        if not _is_internal_frame(frame):
            return (
                frame.f_globals["__name__"],
                frame.f_code.co_name,
                frame.f_lineno,
            )
        frame = frame.f_back
    return "<unknown module>", "<unknown file>", -1


class GenericLogger(BetterABC):
    """通用日志器抽象类

    任何日志器实现本类接口，或通过 :func:`.logger_patch` 修补后，
    即可兼容 melobot 内部所有日志操作（也就可以用于 bot 初始化 :meth:`.Bot.__init__`）
    """

    # pylint: disable=duplicate-code

    @abstractmethod
    def debug(self, msg: object) -> None:
        """`debug` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def info(self, msg: object) -> None:
        """`info` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def warning(self, msg: object) -> None:
        """`warning` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def error(self, msg: object) -> None:
        """`error` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def critical(self, msg: object) -> None:
        """`critical` 级别日志"""
        raise NotImplementedError

    @abstractmethod
    def exception(self, msg: object) -> None:
        """记录异常信息的日志"""
        raise NotImplementedError

    @abstractmethod
    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
    ) -> None:
        """通用懒惰日志方法

        :param msg: 日志消息，可使用 %s 指定稍后填充的参数
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        :param with_exc: 是否记录异常栈信息
        """
        raise NotImplementedError

    @abstractmethod
    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
    ) -> None:
        """通用记录对象日志方法

        :param msg: 附加的日志消息，可使用 %s 指定稍后填充的参数
        :param obj: 需要被日志记录的对象
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        """
        raise NotImplementedError


@singleton
class NullLogger(_Logger, GenericLogger):
    def __init__(self) -> None:
        super().__init__("__MELO_EMPTYLOGGER__", CRITICAL)
        self.addHandler(logging.NullHandler())

    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
    ) -> None:
        return

    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
    ) -> None:
        return


class Logger(_Logger, GenericLogger):
    """melobot 内置日志器

    推荐使用的日志器。实现了 :class:`GenericLogger` 接口，因此可以用于 melobot 内部日志记录。

    `debug`, `info`, `warning`, `error`, `critical`, `exception`
    等接口与 :class:`logging.Logger` 用法完全一致
    """

    __instances__: dict[str, "Logger"] = {}

    def __new__(cls, name: str = "melobot", /, *args: Any, **kwargs: Any) -> Logger:
        if name in Logger.__instances__:
            return Logger.__instances__[name]
        obj = super().__new__(cls)
        Logger.__instances__[name] = obj
        return obj

    def __init__(
        self,
        name: str = "melobot",
        /,
        level: LogLevel = LogLevel.INFO,
        file_level: LogLevel = LogLevel.DEBUG,
        to_console: bool = True,
        to_dir: str | None = None,
        add_tag: bool = False,
        legacy: bool = False,
        yellow_warn: bool = True,
        red_error: bool = True,
        two_stream: bool = False,
    ) -> None:
        """初始化日志器

        :param name: 日志器的名称（唯一）
        :param level: 日志等级
        :param file_level: 日志文件的日志等级
        :param to_console: 是否输出到控制台
        :param to_dir: 保存日志文件的目录，为空则不保存文件
        :param add_tag: 记录日志时是否标识日志器名称
        :param legacy: 记录日志时是否使用传统样式（不对日志内容进行自动高亮，而是使用日志等级的五色）
        :param yellow_warn:
            记录 `LogLevel.WARN` 级别时，是否将日志内容着色为黄色。
            `legacy` 选项为 `True` 时此参数无效

        :param red_error:
            记录 `LogLevel.ERROR` 及以上级别时，是否将日志内容着色为红色。
            `legacy` 选项为 `True` 时此参数无效

        :param two_stream: 当使用记录到文件功能时，是否分离“常规日志”和“错误日志”到不同的文件
        """
        if (
            hasattr(self, "_built")
            and self._built  # pylint: disable=access-member-before-definition
        ):
            return

        super().__init__(name, LogLevel.DEBUG)
        self._handler_arr: list[logging.Handler] = []
        self._no_tag = not add_tag
        self._filter = _MeloLogFilter(name, yellow_warn, red_error, legacy)

        if to_console:
            con_handler = self._add_console_handler()
            con_handler.setLevel(level)

        if to_dir is None:
            pass
        elif not two_stream:
            self._add_file_handler(to_dir, name, file_level)
        else:
            normal_handler = self._add_file_handler(to_dir, f"{name}.out", file_level)
            normal_handler.addFilter(_NoErrFilter(name))
            self._add_file_handler(to_dir, f"{name}.err", max(file_level, LogLevel.ERROR))

        self._built: bool = True

    def _add_console_handler(self) -> logging.Handler:
        fmt = self._console_fmt(self.name, self._no_tag)
        handler = self._console_handler(fmt)
        handler.addFilter(self._filter)

        self.addHandler(handler)
        self._handler_arr.append(handler)
        return handler

    def _add_file_handler(
        self, log_dir: str, name: str, level: LogLevel = LogLevel.DEBUG
    ) -> logging.Handler:
        fmt = self._file_fmt(self.name, self._no_tag)
        handler = self._file_handler(fmt, log_dir, name, level)
        handler.addFilter(self._filter)

        self.addHandler(handler)
        self._handler_arr.append(handler)
        return handler

    @staticmethod
    def _make_fmt_nocache(fmt: logging.Formatter) -> None:
        _origin_format = fmt.format

        def nocache_format(record: logging.LogRecord) -> str:
            record.exc_text = None
            return _origin_format(record)

        fmt.format = nocache_format  # type: ignore[method-assign]

    @staticmethod
    def _console_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        fmt_arr = [
            "%(cyan)s%(asctime)s.%(msecs)03d%(reset)s",
            "%(log_color)s%(levelname)-7s%(reset)s",
            "%(blue)s%(mod_name)s%(reset)s:%(blue)s%(func_name)s%(reset)s"
            + ":%(cyan)s%(func_lineno)d%(reset)s",
        ]
        if not no_tag:
            fmt_arr.insert(0, f"%(purple)s{name}%(reset)s")
        fmt_s = " | ".join(fmt_arr)
        msg_str_f = "%(log_color)s%(legacy_msg_str)s%(reset)s%(colored_msg_str)s"
        obj_str_f = "%(log_color)s%(legacy_obj)s%(reset)s%(colored_obj)s"
        fmt_s += f" - {msg_str_f}{obj_str_f}"

        fmt = colorlog.ColoredFormatter(fmt_s, datefmt="%Y-%m-%d %H:%M:%S", reset=True)
        fmt.default_msec_format = "%s.%03d"
        fmt.formatException = lambda exc_info: "".join(  # type: ignore[assignment]
            _EXC_FORMATTER.format_exception(*exc_info)
        )
        Logger._make_fmt_nocache(fmt)

        # 未知的 mypy bug 推断 fmt 为 Any 类型...
        return fmt  # type: ignore[no-any-return]

    @staticmethod
    def _console_handler(fmt: logging.Formatter) -> logging.Handler:
        handler = logging.StreamHandler()
        handler.setFormatter(fmt)
        return handler

    @staticmethod
    def _file_fmt(name: str, no_tag: bool = False) -> logging.Formatter:
        fmt_arr = [
            "%(asctime)s.%(msecs)03d",
            "%(levelname)-7s",
            "%(mod_name)s:%(func_name)s:%(func_lineno)d",
        ]
        if not no_tag:
            fmt_arr.insert(0, name)
        fmt_s = " | ".join(fmt_arr)
        fmt_s += " - %(msg_str)s%(obj)s"

        fmt = logging.Formatter(
            fmt=fmt_s,
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        fmt.default_msec_format = "%s.%03d"
        fmt.formatException = lambda exc_info: "".join(  # type: ignore
            _NO_COLOR_EXC_FORMATTER.format_exception(*exc_info)
        )
        Logger._make_fmt_nocache(fmt)

        return fmt

    @staticmethod
    def _file_handler(
        fmt: logging.Formatter, log_dir: str, name: str, level: LogLevel
    ) -> logging.Handler:
        if not os.path.exists(log_dir):
            os.mkdir(log_dir)

        handler = logging.handlers.RotatingFileHandler(
            filename=os.path.join(log_dir, f"{name}.log"),
            maxBytes=1024 * 1024,
            backupCount=10,
            encoding="UTF-8",
        )
        handler.setLevel(level)
        handler.setFormatter(fmt)
        return handler

    def set_level(self, level: LogLevel) -> None:  # type: ignore[override]
        """设置日志等级

        日志等级自动应用于包含的所有 handler（但输出日志到文件的 handler 除外）

        :param level: 日志等级
        """
        super().setLevel(level)
        for handler in self._handler_arr:
            if not isinstance(handler, logging.handlers.RotatingFileHandler):
                handler.setLevel(level)

    def findCaller(
        self, stack_info: bool = False, stacklevel: int = 1
    ) -> tuple[str, int, str, str | None]:
        f = currentframe()
        if f is None:
            return "<unknown file>", 0, "<unknown function>", "<unknown stackinfo>"

        while stacklevel > 0:
            next_f = f.f_back
            if next_f is None:
                break
            f = next_f
            if not _is_internal_frame(f):
                stacklevel -= 1
        co = f.f_code
        sinfo = None

        if stack_info:
            with io.StringIO() as sio:
                sio.write("Stack (most recent call last):\n")
                traceback.print_stack(f, file=sio)
                sinfo = sio.getvalue()
                if sinfo[-1] == "\n":
                    sinfo = sinfo[:-1]

        assert isinstance(f.f_lineno, int)
        return co.co_filename, f.f_lineno, co.co_name, sinfo

    def generic_lazy(
        self,
        msg: str,
        *arg_getters: Callable[[], str],
        level: LogLevel,
        with_exc: bool = False,
    ) -> None:
        """懒惰日志方法

        :param msg: 日志消息，可使用 %s 指定稍后填充的参数
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        :param with_exc: 是否记录异常栈信息
        """
        if not self.isEnabledFor(level):
            return
        exc = sys.exc_info() if with_exc else None
        self._log(level, msg, tuple(g() for g in arg_getters), exc_info=exc)

    def generic_obj(
        self,
        msg: str,
        obj: T,
        *arg_getters: Callable[[], str],
        level: LogLevel = LogLevel.INFO,
    ) -> None:
        """记录对象的日志方法

        :param msg: 附加的日志消息，可使用 %s 指定稍后填充的参数
        :param obj: 需要被日志记录的对象
        :param arg_getters: 填充消息 %s 位置的填充函数
        :param level: 日志等级
        """
        if isinstance(self, Logger):
            with self._filter.on_obj(obj):
                self.generic_lazy(msg + "\n", *arg_getters, level=level)
        else:
            _getters = arg_getters + (lambda: _get_rich_object(obj)[1],)
            self.generic_lazy(msg + "\n%s", *_getters, level=level)



from __future__ import annotations

import asyncio
from asyncio import create_task
from contextlib import contextmanager
from dataclasses import dataclass, field
from time import time_ns

from typing_extensions import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Coroutine,
    Generator,
    Generic,
    Hashable,
    Literal,
    LiteralString,
    Self,
    Sequence,
    TypeVar,
    cast,
)

from ..ctx import ActionManualSignalCtx, Context, LoggerCtx
from ..exceptions import AdapterError
from ..io.base import AbstractOutSource
from ..log.base import LogLevel
from ..mixin import AttrReprMixin, FlagMixin
from ..typ.base import T
from ..typ.cls import abstractattr
from ..utils.base import to_coro
from ..utils.common import get_id
from .content import Content

if TYPE_CHECKING:
    from .base import AbstractEchoFactory, AbstractOutputFactory


class Event(AttrReprMixin, FlagMixin):
    """事件基类

    :ivar typing.LiteralString protocol: 遵循的协议，为空则协议无关
    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar typing.Sequence[Content] contents: 附加的通用内容序列
    """

    def __init__(
        self,
        protocol: LiteralString,
        time: float = -1,
        id: str = "",
        scope: Hashable | None = None,
        contents: Sequence[Content] | None = None,
    ) -> None:
        super().__init__()

        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.contents = contents if contents else ()
        self.scope = scope

        self.spread: bool = True


EventT = TypeVar("EventT", bound=Event)


class TextEvent(Event):
    """文本事件类

    :ivar str text: 文本内容
    """

    text: str = abstractattr()
    textlines: list[str] = abstractattr()


class Action(AttrReprMixin, FlagMixin):
    """行为基类

    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.LiteralString | None protocol: 遵循的协议，为空则协议无关
    :ivar typing.Sequence[Content] contents: 附加的通用内容序列
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar Event | None trigger: 触发该行为的事件，为空表明不由事件触发
    """

    def __init__(
        self,
        time: float = -1,
        id: str = "",
        protocol: LiteralString | None = None,
        scope: Hashable | None = None,
        contents: Sequence[Content] | None = None,
        trigger: Event | None = None,
    ) -> None:
        super().__init__()
        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.contents = contents if contents else ()
        self.scope = scope
        self.trigger = trigger


class Echo(AttrReprMixin, FlagMixin):
    """回应基类

    :ivar float time: 时间戳
    :ivar str id: id 标识
    :ivar typing.LiteralString | None protocol: 遵循的协议，为空则协议无关
    :ivar typing.Hashable | None scope: 所在的域，可空
    :ivar bool ok: 回应是否成功
    :ivar int status: 回应状态码
    :ivar str prompt: 回应提示语
    :ivar Any data: 回应数据
    """

    def __init__(
        self,
        time: float = -1,
        id: str = "",
        protocol: LiteralString | None = None,
        scope: Hashable | None = None,
        ok: bool = True,
        status: int = 0,
        prompt: str = "",
        data: Any = None,
    ) -> None:
        super().__init__()
        self.time = time_ns() / 1e9 if time == -1 else time
        self.id = get_id() if id == "" else id
        self.protocol = protocol
        self.scope = scope
        self.ok = ok
        self.status = status
        self.prompt = prompt
        self.data = data


ActionT = TypeVar("ActionT", bound=Action)
EchoT = TypeVar("EchoT", bound=Echo)


ActionRetT = TypeVar("ActionRetT", bound=Echo | None)


class ActionHandle(Generic[ActionRetT]):
    """行为操作句柄

    :ivar Action action: 操作包含的行为对象
    :ivar typing.Literal["PENDING", "EXECUTING", "FINISHED"] status: 操作的状态。分别对应：未执行、执行中、执行完成
    :ivar AbstractOutSource out_src: 执行操作的输出源对象
    """

    def __init__(
        self,
        action: Action,
        out_src: AbstractOutSource,
        output_factory: "AbstractOutputFactory",
        echo_factory: "AbstractEchoFactory",
    ) -> None:
        self.action = action
        self.status: Literal["PENDING", "EXECUTING", "FINISHED"] = "PENDING"
        self.out_src = out_src

        self._echo: ActionRetT
        self._output_factory = output_factory
        self._echo_factory = echo_factory
        self._done = asyncio.Event()

        if not ActionManualSignalCtx().try_get():
            self.execute()

    async def _wait(self) -> ActionRetT:
        await self._done.wait()
        return self._echo

    def __await__(self) -> Generator[Any, Any, ActionRetT]:
        """本对象实现 __await__ 接口，因此可等待。返回本操作对应的回应数据"""
        return self._wait().__await__()

    async def _execute(self) -> None:
        try:
            output_packet = await self._output_factory.create(self.action)
            echo_packet = await self.out_src.output(output_packet)
            self._echo = cast(ActionRetT, await self._echo_factory.create(echo_packet))
            self.status = "FINISHED"
            self._done.set()
        except Exception:
            logger = LoggerCtx().get()
            handle_name = (
                f"{self.__class__.__module__}.{self.action.__class__.__qualname__}"
            )
            logger.exception(f"行为句柄 {handle_name} 执行时出现异常")
            logger.generic_obj("异常点局部变量", self.__dict__, level=LogLevel.ERROR)

    def execute(self) -> Self:
        if self.status != "PENDING":
            raise AdapterError("行为操作正在执行或执行完毕，不应该再执行")

        self.status = "EXECUTING"
        create_task(self._execute())
        return self


@dataclass
class _ChainStep:
    coros: Sequence[Coroutine]
    ret_when: Literal["FIRST_COMPLETED", "FIRST_EXCEPTION", "ALL_COMPLETED"]
    next: _ChainStep | None = None


@dataclass(kw_only=True)
class _ChainCtxStep(_ChainStep):  # type: ignore[override]
    ctx_var: Context
    ctx_val: Any
    coros: Sequence[Coroutine] = field(default_factory=tuple)
    ret_when: Literal["ALL_COMPLETED"] = "ALL_COMPLETED"


class ActionChain:
    """行为链"""

    def __init__(self) -> None:
        self._chain: list[_ChainStep] = []

    def _add_step(self, step: _ChainStep) -> None:
        if len(self._chain):
            self._chain[-1].next = step
        self._chain.append(step)

    def unfold(self, ctx: Context[T], val: T) -> Self:
        """指定链后续的步骤中在 `ctx` 类别的上下文中执行，上下文值为 `val`

        :param ctx: 上下文类别
        :param val: 上下文的值
        :return: 自身，因此支持链式调用
        """
        self._add_step(_ChainCtxStep(ctx_var=ctx, ctx_val=val))
        return self

    async def _exec_handle(self, handles: Awaitable[tuple[ActionHandle, ...]]) -> None:
        _handles = await handles
        for handle in _handles:
            handle.execute()
        if len(_handles):
            await asyncio.wait(map(create_task, map(to_coro, _handles)))

    def add(
        self,
        *handles: Awaitable[tuple[ActionHandle, ...]],
        ret_when: Literal[
            "FIRST_COMPLETED", "FIRST_EXCEPTION", "ALL_COMPLETED"
        ] = "ALL_COMPLETED",
    ) -> Self:
        """在链的步骤中添加一组行为

        :param handles: 返回行为操作句柄的可等待对象
        :param ret_when: 指定这一组行为的等待模式
        :return: 自身，因此支持链式调用
        """
        coros = tuple(self._exec_handle(hs) for hs in handles)
        self._add_step(_ChainStep(coros, ret_when))
        return self

    def sleep(self, interval: float) -> Self:
        """
        在链的步骤中添加指定时长的等待

        :param interval: 等待时长
        :return: 自身，因此支持链式调用
        """
        coros = (asyncio.sleep(interval),)
        self._add_step(_ChainStep(coros, ret_when="ALL_COMPLETED"))
        return self

    async def _start(self, step: _ChainStep) -> None:
        if isinstance(step, _ChainCtxStep):
            with step.ctx_var.unfold(step.ctx_val):
                if step.next:
                    await self._start(step.next)
            return

        if len(step.coros):
            await asyncio.wait(map(create_task, step.coros), return_when=step.ret_when)
        if step.next:
            await self._start(step.next)

    async def run(self) -> None:
        """顺序执行链的所有步骤"""
        await self._start(self._chain[0])


@contextmanager
def open_chain() -> Generator[ActionChain, None, None]:
    """创建行为链的上下文管理器

    :yield: 行为链对象
    """
    with ActionManualSignalCtx().unfold(True):
        yield ActionChain()



from ..adapter.model import Event as _Event
from ..ctx import BotCtx as _BotCtx
from ..ctx import FlowCtx as _FlowCtx
from ..ctx import FlowRecord, FlowRecordStage, FlowStore
from .base import (
    Flow,
    FlowNode,
    block,
    bypass,
    flow_to,
    nextn,
    no_deps_node,
    node,
    rewind,
    stop,
)
from .register import (
    FlowDecorator,
    GetParseArgs,
    on_command,
    on_contain_match,
    on_end_match,
    on_event,
    on_full_match,
    on_regex_match,
    on_start_match,
    on_text,
)


def get_flow_records() -> tuple[FlowRecord, ...]:
    """获取当前上下文中的流记录

    :return: 流记录
    """
    return _FlowCtx().get_records()


def get_flow_store() -> FlowStore:
    """获取当前上下文中的流存储

    :return: 流存储
    """
    return _FlowCtx().get_store()


def get_event() -> _Event:
    """获取当前上下文中的事件

    :return: 事件
    """
    return _FlowCtx().get_event()


def try_get_event() -> _Event | None:
    """尝试获取当前上下文中的事件

    :return: 事件或空
    """
    return _FlowCtx().try_get_event()



from .base import (
    AbstractInSource,
    AbstractIOSource,
    AbstractOutSource,
    AbstractSource,
    EchoPacket,
    EchoPacketT,
    InOrOutSourceT,
    InPacket,
    InPacketT,
    InSourceT,
    IOSourceT,
    OutPacket,
    OutPacketT,
    OutSourceT,
    SourceLifeSpan,
)



from .atool import async_at, async_interval, async_later, call_at, call_later
from .base import async_guard, to_async, to_coro, to_sync
from .common import (
    DeprecatedLoader,
    RWContext,
    deprecate_warn,
    deprecated,
    get_id,
    get_obj_name,
    singleton,
)
from .deco import cooldown, if_not, lock, semaphore, speedlimit, timelimit, unfold_ctx



from __future__ import annotations

from dataclasses import dataclass

from typing_extensions import TYPE_CHECKING, Any, Callable, Generic, final

from ..adapter.model import Event, EventT
from ..typ.cls import BetterABC

if TYPE_CHECKING:
    from .base import Session


@dataclass
class CompareInfo(Generic[EventT]):
    """用于会话判断的信息"""

    session: "Session"
    old_event: EventT
    new_event: EventT


class Rule(BetterABC, Generic[EventT]):
    """会话规则

    对于更复杂的情况，可以继承此类，在子类中结合状态信息实现更动态的会话判断
    """

    @staticmethod
    @final
    def new(meth: Callable[[EventT, EventT], bool]) -> Rule[EventT]:
        """从可调用对象创建一个新的会话规则对象

        :param meth: 用于会话判断的可调用对象
        :return: 会话规则
        """
        return _CustomRule[EventT](meth)

    async def compare(self, e1: EventT, e2: EventT) -> bool:
        """会话判断的方法

        抽象方法，和 :meth:`compare_with` 二选一实现

        :param e1: 某一已存在的会话中的事件
        :param e2: 待判断的事件
        :return: 为 `True` 则在一个会话中，反之亦然
        """
        raise NotImplementedError

    async def compare_with(self, info: CompareInfo[EventT]) -> bool:
        """会话判断的方法

        抽象方法，和 :meth:`compare` 二选一实现

        :param info: 用于会话判断的信息
        :return: 为 `True` 则在一个会话中，反之亦然
        """
        raise NotImplementedError

    async def __default_compare_with__(self, info: CompareInfo[EventT]) -> bool:
        return await self.compare(info.old_event, info.new_event)

    def __init_subclass__(cls, *args: Any, **kwargs: Any) -> None:
        super().__init_subclass__(*args, **kwargs)
        has_comp = cls.compare is not Rule.compare
        has_comp_with = cls.compare_with is not Rule.compare_with

        if has_comp_with:
            return
        if has_comp:
            cls.compare_with = cls.__default_compare_with__  # type: ignore[method-assign]
            return

        raise TypeError(
            f"{cls.__name__} 类必须实现抽象方法"
            f"'{cls.compare.__name__}', '{cls.compare_with.__name__}' 其中之一"
        ) from None


class _CustomRule(Rule[EventT]):
    # pylint: disable=abstract-method
    def __init__(self, meth: Callable[[EventT, EventT], bool]) -> None:
        super().__init__()
        self.meth = meth

    async def compare(self, e1: EventT, e2: EventT) -> bool:
        return self.meth(e1, e2)


class DefaultRule(Rule[Event]):
    """传统的会话规则

    判断事件的 `scope` 是否相同
    """

    async def compare(self, e1: Event, e2: Event) -> bool:
        return e1.scope == e2.scope



import asyncio
from asyncio import Queue, create_task

from melobot.adapter.generic import send_text
from melobot.bot import Bot
from melobot.handle import Flow, node
from melobot.log import GenericLogger
from melobot.plugin import PluginPlanner
from melobot.protocols.onebot.v11.adapter.base import Adapter
from melobot.protocols.onebot.v11.adapter.event import MessageEvent
from melobot.protocols.onebot.v11.io.base import BaseIOSource
from melobot.protocols.onebot.v11.io.packet import EchoPacket, InPacket, OutPacket
from tests.base import *

_TEST_EVENT_DICT = {
    "time": 1725292489,
    "self_id": 123456,
    "post_type": "message",
    "message_type": "group",
    "sub_type": "normal",
    "sender": {
        "age": 0,
        "nickname": "这是一个群昵称",
        "sex": "unknown",
        "user_id": 1574260633,
        "area": "",
        "card": "",
        "level": "",
        "role": "member",
        "title": "",
    },
    "message_id": -1234567890,
    "font": 0,
    "message": "",
    "user_id": 1574260633,
    "anonymous": None,
    "group_id": 535705163,
    "raw_message": "",
}

_SUCCESS_SIGNAL = asyncio.Event()


class TempIO(BaseIOSource):
    def __init__(self) -> None:
        super().__init__(1)
        self.queue = Queue()
        self.queue.put_nowait(InPacket(data=_TEST_EVENT_DICT))

    async def open(self) -> None:
        pass

    def opened(self) -> bool:
        return True

    async def close(self) -> None:
        pass

    async def input(self) -> InPacket:
        return await self.queue.get()

    async def output(self, packet: OutPacket) -> EchoPacket:
        if packet.echo_id is None:
            return EchoPacket(noecho=True)
        return EchoPacket(
            data={"data": {"message_id": 123456}, "status": "ok", "retcode": 0},
            action_type="send_msg",
        )


@node
async def process(adapter: Adapter, event: MessageEvent, logger: GenericLogger) -> None:
    assert isinstance(event, MessageEvent)

    pending = await adapter.with_echo(send_text)("generic send test")
    assert (await pending[0]).data["message_id"] == 123456

    pending = await adapter.__send_media__("test.bmp", url="https://example.com/test.bmp")
    assert pending[0].action.params["message"][0]["type"] == "share"
    pending = await adapter.__send_media__("test.bmp", raw=b"123")
    assert pending[0].action.params["message"][0]["type"] == "text"

    pending = await adapter.__send_image__("test.jpg", url="https://example.com/test.jpg")
    assert pending[0].action.params["message"][0]["type"] == "image"

    pending = await adapter.__send_file__("test.txt", path="/home/abc/test.txt")
    assert pending[0].action.params["message"][0]["type"] == "text"

    pending = await adapter.__send_refer__(event)
    assert pending[0].action.params["message"][0]["type"] == "reply"

    pending = await adapter.__send_resource__("123456", "https://example.com/test.jpg")
    assert pending[0].action.params["message"][0]["type"] == "share"

    logger.info("adapter main event process ok")
    _SUCCESS_SIGNAL.set()


async def after_bot_started(bot: Bot):
    adapter = next(iter(bot.adapters.values()))
    pending = await adapter.with_echo(adapter.send_custom)("Hello World!", user_id=12345)
    data = (await pending[0]).data
    mid = data["message_id"]
    assert mid == 123456
    await bot.close()


async def test_adapter_base():
    mbot = Bot("test_adapter_base")
    mbot.add_io(TempIO())
    mbot.add_adapter(Adapter())

    flow = Flow("test_adapter_base", [process])
    mbot.load_plugin(PluginPlanner("1.0.0", flows=[flow]))

    mbot.on_started(after_bot_started)
    create_task(mbot.core_run())
    await mbot._rip_signal.wait()
    await _SUCCESS_SIGNAL.wait()



from enum import Enum

from typing_extensions import Any, Callable, Sequence

from .base import AsyncCallable


class VoidType(Enum):
    """空类型，需要区别于 `None` 时使用

    .. code:: python

        # 有些时候 `None` 也是合法值，因此需要一个额外的哨兵值：
        def foo(val: Any | VoidType = VoidType.VOID) -> None:
            ...
    """

    VOID = type("_VOID", (), {})


class LogicMode(Enum):
    """逻辑模式枚举类型"""

    AND = 1
    OR = 2
    NOT = 3
    XOR = 4

    @classmethod
    def calc(cls, logic: "LogicMode", v1: Any, v2: Any = None) -> bool:
        """将两个值使用指定逻辑模式运算

        :param logic: 逻辑模式
        :param v1: 值 1
        :param v2: 值 2
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            return (v1 and v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            return (v1 or v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not v1
        return (v1 ^ v2) if v2 is not None else bool(v1)  # type: ignore[no-any-return]

    @classmethod
    def short_calc(
        cls, logic: "LogicMode", v1: Callable[[], Any], v2: Callable[[], Any] | None
    ) -> bool:
        """与 :func:`calc` 功能类似，但运算支持短路

        :param logic: 逻辑模式
        :param v1: 生成值 1 的可调用对象
        :param v2: 生成值 2 的可调用对象
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            return (v1() and v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            return (v1() or v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not v1()
        return (v1() ^ v2()) if v2 is not None else bool(v1())  # type: ignore[no-any-return]

    @classmethod
    async def async_short_calc(
        cls,
        logic: "LogicMode",
        v1: AsyncCallable[[], Any],
        v2: AsyncCallable[[], Any] | None,
    ) -> bool:
        """与 :func:`short_calc` 功能类似，但运算支持异步

        :param logic: 逻辑模式
        :param v1: 生成值 1 的异步可调用对象
        :param v2: 生成值 2 的异步可调用对象
        :return: 布尔值
        """
        if logic == LogicMode.AND:
            res = (await v1() and await v2()) if v2 is not None else bool(await v1())
            return res  # type: ignore[no-any-return]
        if logic == LogicMode.OR:
            res = (await v1() or await v2()) if v2 is not None else bool(await v1())
            return res  # type: ignore[no-any-return]
        if logic == LogicMode.NOT:
            return not await v1()
        res = (await v1() ^ await v2()) if v2 is not None else bool(await v1())
        return res  # type: ignore[no-any-return]

    @classmethod
    def seq_calc(cls, logic: "LogicMode", values: list[Any]) -> bool:
        """使用指定的逻辑模式，对值序列进行运算

        .. code:: python

            # 操作等价与：True and False and True
            LogicMode.seq_calc(LogicMode.AND, [True, False, True])

        :param logic: 逻辑模式
        :param values: 值序列
        :return: 布尔值
        """
        if len(values) <= 0:
            return False
        if len(values) <= 1:
            return bool(values[0])

        idx = 0
        res: bool
        while idx < len(values):
            if idx == 0:
                res = cls.calc(logic, values[idx], values[idx + 1])
                idx += 1
            else:
                res = cls.calc(logic, res, values[idx])
            idx += 1
        return res

    @classmethod
    def short_seq_calc(
        cls, logic: "LogicMode", getters: Sequence[Callable[[], Any]]
    ) -> bool:
        """与 :func:`seq_calc` 功能类似，但运算支持短路

        :param logic: 逻辑模式
        :param getters: 一组获取值的可调用对象
        :return: 布尔值
        """
        if len(getters) <= 0:
            return False
        if len(getters) <= 1:
            return bool(getters[0]())

        idx = 0
        res: bool
        while idx < len(getters):
            if idx == 0:
                res = cls.short_calc(logic, getters[idx], getters[idx + 1])
                idx += 1
            else:
                res = cls.short_calc(logic, lambda: res, getters[idx])
            idx += 1
        return res

    @classmethod
    async def async_short_seq_calc(
        cls, logic: "LogicMode", getters: Sequence[AsyncCallable[[], Any]]
    ) -> bool:
        """与 :func:`short_seq_calc` 功能类似，但运算支持异步

        :param logic: 逻辑模式
        :param getters: 一组获取值的异步可调用对象
        :return: 布尔值
        """
        if len(getters) <= 0:
            return False
        if len(getters) <= 1:
            return bool(await getters[0]())

        idx = 0
        res: bool
        while idx < len(getters):
            if idx == 0:
                res = await cls.async_short_calc(logic, getters[idx], getters[idx + 1])
                idx += 1
            else:

                async def res_getter() -> bool:
                    return res

                res = await cls.async_short_calc(logic, res_getter, getters[idx])
            idx += 1
        return res



import json

from typing_extensions import Any, Iterable, Literal, Optional, TypedDict

from melobot.adapter import Action as RootAction
from melobot.handle import try_get_event

from ..const import PROTOCOL_IDENTIFIER
from .segment import NodeSegment, Segment, TextSegment


class Action(RootAction):
    def __init__(self, type: str, params: dict[str, Any]) -> None:
        self.time: int

        super().__init__(protocol=PROTOCOL_IDENTIFIER, trigger=try_get_event())

        self.type = type
        self.params = params
        self.need_echo = False

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(type={self.type}, echo={self.need_echo})"

    def set_echo(self, status: bool) -> None:
        self.need_echo = status

    def extract(self) -> dict[str, Any]:
        obj = {
            "action": self.type,
            "params": self.params,
        }
        if self.need_echo:
            obj["echo"] = self.id
        return obj

    def flatten(self) -> str:
        return json.dumps(self.extract(), ensure_ascii=False)


def msgs_to_dicts(
    msgs: str | Segment | Iterable[Segment] | dict | Iterable[dict],
) -> list[dict]:
    if isinstance(msgs, str):
        return [TextSegment(msgs).to_dict(force_str=True)]
    if isinstance(msgs, Segment):
        return [msgs.to_dict(force_str=True)]
    if isinstance(msgs, dict):
        return [msgs]

    return [
        msg.to_dict(force_str=True) if isinstance(msg, Segment) else msg for msg in msgs
    ]


class SendMsgAction(Action):
    def __init__(
        self,
        msgs: str | Segment | Iterable[Segment] | dict | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> None:
        type = "send_msg"
        _msgs = msgs_to_dicts(msgs)
        if group_id is None:
            assert user_id is not None, "group_id 为空时，user_id 必须为非空"
            params = {
                "message_type": "private",
                "user_id": user_id,
                "message": _msgs,
                "auto_escape": False,
            }
        else:
            params = {
                "message_type": "group",
                "group_id": group_id,
                "message": _msgs,
                "auto_escape": False,
            }
        super().__init__(type, params)


class SendForwardMsgAction(Action):
    def __init__(
        self,
        msgs: Iterable[NodeSegment] | Iterable[dict],
        user_id: Optional[int] = None,
        group_id: Optional[int] = None,
    ) -> None:
        _msgs = msgs_to_dicts(msgs)
        if group_id is None:
            type = "send_private_forward_msg"
            params = {"user_id": user_id, "messages": _msgs, "auto_escape": False}
        else:
            type = "send_group_forward_msg"
            params = {"group_id": group_id, "messages": _msgs, "auto_escape": False}
        super().__init__(type, params)


class DeleteMsgAction(Action):
    def __init__(self, msg_id: int) -> None:
        type = "delete_msg"
        params = {
            "message_id": msg_id,
        }
        super().__init__(type, params)


class GetMsgAction(Action):
    def __init__(self, msg_id: int) -> None:
        type = "get_msg"
        params = {"message_id": msg_id}
        super().__init__(type, params)


class GetForwardMsgAction(Action):
    def __init__(self, forward_id: str) -> None:
        type = "get_forward_msg"
        params = {"id": forward_id}
        super().__init__(type, params)


class SendLikeAction(Action):
    def __init__(self, user_id: int, times: int = 1) -> None:
        type = "send_like"
        params = {"user_id": user_id, "times": times}
        super().__init__(type, params)


class SetGroupKickAction(Action):
    def __init__(self, group_id: int, user_id: int, later_reject: bool = False) -> None:
        type = "set_group_kick"
        params = {
            "group_id": group_id,
            "user_id": user_id,
            "reject_add_request": later_reject,
        }
        super().__init__(type, params)


class SetGroupBanAction(Action):
    def __init__(self, group_id: int, user_id: int, duration: int = 30 * 60) -> None:
        type = "set_group_ban"
        params = {
            "group_id": group_id,
            "user_id": user_id,
            "duration": duration,
        }
        super().__init__(type, params)


class SetGroupAnonymousBanAction(Action):

    class AnonymousDict(TypedDict):
        id: int
        name: str
        flag: str

    def __init__(
        self,
        group_id: int,
        anonymous: AnonymousDict,
        anonymous_flag: str,
        duration: int = 30 * 60,
    ) -> None:
        type = "set_group_anonymous_ban"
        params = {
            "group_id": group_id,
            "anonymous": anonymous,
            "anonymous_flag": anonymous_flag,
            "duration": duration,
        }
        super().__init__(type, params)


class SetGroupWholeBanAction(Action):
    def __init__(self, group_id: int, enable: bool = True) -> None:
        type = "set_group_whole_ban"
        params = {"group_id": group_id, "enable": enable}
        super().__init__(type, params)


class SetGroupAdminAction(Action):
    def __init__(self, group_id: int, user_id: int, enable: bool = True) -> None:
        type = "set_group_admin"
        params = {"group_id": group_id, "user_id": user_id, "enable": enable}
        super().__init__(type, params)


class SetGroupAnonymousAction(Action):
    def __init__(self, group_id: int, enable: bool = True) -> None:
        type = "set_group_anonymous"
        params = {"group_id": group_id, "enable": enable}
        super().__init__(type, params)


class SetGroupCardAction(Action):
    def __init__(self, group_id: int, user_id: int, card: str = "") -> None:
        type = "set_group_card"
        params = {"group_id": group_id, "user_id": user_id, "card": card}
        super().__init__(type, params)


class SetGroupNameAction(Action):
    def __init__(self, group_id: int, name: str) -> None:
        type = "set_group_name"
        params = {"group_id": group_id, "group_name": name}
        super().__init__(type, params)


class SetGroupLeaveAction(Action):
    def __init__(self, group_id: int, is_dismiss: bool = False) -> None:
        type = "set_group_leave"
        params = {"group_id": group_id, "is_dismiss": is_dismiss}
        super().__init__(type, params)


class SetGroupSpecialTitleAction(Action):
    def __init__(
        self, group_id: int, user_id: int, title: str = "", duration: int = -1
    ) -> None:
        type = "set_group_special_title"
        params = {
            "group_id": group_id,
            "user_id": user_id,
            "special_title": title,
            "duration": duration,
        }
        super().__init__(type, params)


class SetFriendAddRequestAction(Action):
    def __init__(self, add_flag: str, approve: bool = True, remark: str = "") -> None:
        type = "set_friend_add_request"
        params = {"flag": add_flag, "approve": approve, "remark": remark}
        super().__init__(type, params)


class SetGroupAddRequestAction(Action):
    def __init__(
        self,
        add_flag: str,
        add_type: Literal["add", "invite"],
        approve: bool = True,
        reason: str = "",
    ) -> None:
        type = "set_group_add_request"
        params = {
            "flag": add_flag,
            "sub_type": add_type,
            "approve": approve,
        }
        if reason:
            self.params["reason"] = reason
        super().__init__(type, params)


class GetLoginInfoAction(Action):
    def __init__(self) -> None:
        type = "get_login_info"
        params: dict = {}
        super().__init__(type, params)


class GetStrangerInfoAction(Action):
    def __init__(self, user_id: int, no_cache: bool = False) -> None:
        type = "get_stranger_info"
        params = {"user_id": user_id, "no_cache": no_cache}
        super().__init__(type, params)


class GetFriendlistAction(Action):
    def __init__(self) -> None:
        type = "get_friend_list"
        params: dict = {}
        super().__init__(type, params)


class GetGroupInfoAction(Action):
    def __init__(self, group_id: int, no_cache: bool = False) -> None:
        type = "get_group_info"
        params = {"group_id": group_id, "no_cache": no_cache}
        super().__init__(type, params)


class GetGrouplistAction(Action):
    def __init__(self) -> None:
        type = "get_group_list"
        params: dict = {}
        super().__init__(type, params)


class GetGroupMemberInfoAction(Action):
    def __init__(self, group_id: int, user_id: int, no_cache: bool = False) -> None:
        type = "get_group_member_info"
        params = {"group_id": group_id, "user_id": user_id, "no_cache": no_cache}
        super().__init__(type, params)


class GetGroupMemberlistAction(Action):
    def __init__(self, group_id: int) -> None:
        type = "get_group_member_list"
        params = {"group_id": group_id}
        super().__init__(type, params)


class GetGroupHonorInfoAction(Action):
    def __init__(
        self,
        group_id: int,
        type: Literal[
            "talkative", "performer", "legend", "strong_newbie", "emotion", "all"
        ],
    ) -> None:
        _type = "get_group_honor_info"
        params = {"group_id": group_id, "type": type}
        super().__init__(_type, params)


class GetCookiesAction(Action):
    def __init__(self, domain: str = "") -> None:
        type = "get_cookies"
        params = {"domain": domain}
        super().__init__(type, params)


class GetCsrfTokenAction(Action):
    def __init__(self) -> None:
        type = "get_csrf_token"
        params: dict = {}
        super().__init__(type, params)


class GetCredentialsAction(Action):
    def __init__(self, domain: str = "") -> None:
        type = "get_credentials"
        params = {"domain": domain}
        super().__init__(type, params)


class GetRecordAction(Action):
    def __init__(self, file: str, out_format: str) -> None:
        type = "get_record"
        params = {"file": file, "out_format": out_format}
        super().__init__(type, params)


class GetImageAction(Action):
    def __init__(self, file: str) -> None:
        type = "get_image"
        params = {"file": file}
        super().__init__(type, params)


class CanSendImageAction(Action):
    def __init__(self) -> None:
        type = "can_send_image"
        params: dict = {}
        super().__init__(type, params)


class CanSendRecordAction(Action):
    def __init__(self) -> None:
        type = "can_send_record"
        params: dict = {}
        super().__init__(type, params)


class GetStatusAction(Action):
    def __init__(self) -> None:
        type = "get_status"
        params: dict = {}
        super().__init__(type, params)


class GetVersionInfoAction(Action):
    def __init__(self) -> None:
        type = "get_version_info"
        params: dict = {}
        super().__init__(type, params)


class SetRestartAction(Action):
    def __init__(self, delay: int = 0) -> None:
        type = "set_restart"
        params = {"delay": delay}
        super().__init__(type, params)


class CleanCacheAction(Action):
    def __init__(self) -> None:
        type = "clean_cache"
        params: dict = {}
        super().__init__(type, params)
